/*-
# X-BASED BARREL(tm)
#
#  Barrel.c
#
###
#
#  Copyright (c) 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Barrel */
#ifdef WINVER
#include "BarrelP.h"

#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wbarrel.ini"
#endif

#define SECTION "setup"
HBRUSH      hOldBrush, hBrush;
HPEN        hOldPen, hPen;

static const char *faceColorString[MAXCOLORS] =
{
	"0 255 0",
	"255 0 0",
	"255 255 0",
	"255 165 0",
	"0 0 255",
	"0 0 0"
};

static char faceColorChar[MAXCOLORS] =
{'G', 'R', 'Y', 'O', 'B', 'b'};
#else
#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "BarrelP.h"

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void InitializeBarrel(Widget request, Widget renew);
static void ExposeBarrel(Widget renew, XEvent * event, Region region);
static void ResizeBarrel(BarrelWidget w);
static void DestroyBarrel(Widget old);
static Boolean SetValuesBarrel(Widget current, Widget request, Widget renew);
static void QuitBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void HideBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void SelectBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void ReleaseBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void PracticeBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void PracticeBarrelMaybe(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void PracticeBarrel2(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeBarrelMaybe(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeBarrel2(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void GetBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void WriteBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void ClearBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void UndoBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void SolveBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void OrientizeBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void PairsBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void EnterBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void LeaveBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void MoveBarrelTop(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void MoveBarrelLeft(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void MoveBarrelRight(BarrelWidget w, XEvent * event, char **args, int nArgs);
static void MoveBarrelBottom(BarrelWidget w, XEvent * event, char **args, int nArgs);

static char defaultTranslationsBarrel[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>Up: MoveTop()\n\
 <KeyPress>osfUp: MoveTop()\n\
 <KeyPress>KP_Up: MoveTop()\n\
 <KeyPress>KP_8: MoveTop()\n\
 <KeyPress>R8: MoveTop()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <KeyPress>Down: MoveBottom()\n\
 <KeyPress>osfDown: MoveBottom()\n\
 <KeyPress>KP_Down: MoveBottom()\n\
 <KeyPress>KP_2: MoveBottom()\n\
 <KeyPress>R14: MoveBottom()\n\
 <Btn1Down>: Select()\n\
 <Btn1Up>: Release()\n\
 <KeyPress>p: Practice()\n\
 <KeyPress>r: Randomize()\n\
 <Btn3Down>: RandomizeMaybe()\n\
 <Btn3Down>(2+): Randomize2()\n\
 <Btn4Down>: MoveTop()\n\
 <Btn5Down>: MoveBottom()\n\
 <KeyPress>g: Get()\n\
 <KeyPress>w: Write()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>u: Undo()\n\
 <KeyPress>s: Solve()\n\
 <KeyPress>o: Orientize()\n\
 <KeyPress>a: Pairs()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListBarrel[] =
{
	{(char *) "Quit", (XtActionProc) QuitBarrel},
	{(char *) "Hide", (XtActionProc) HideBarrel},
	{(char *) "MoveTop", (XtActionProc) MoveBarrelTop},
	{(char *) "MoveLeft", (XtActionProc) MoveBarrelLeft},
	{(char *) "MoveRight", (XtActionProc) MoveBarrelRight},
	{(char *) "MoveBottom", (XtActionProc) MoveBarrelBottom},
	{(char *) "Select", (XtActionProc) SelectBarrel},
	{(char *) "Release", (XtActionProc) ReleaseBarrel},
	{(char *) "Practice", (XtActionProc) PracticeBarrel},
	{(char *) "PracticeMaybe", (XtActionProc) PracticeBarrelMaybe},
	{(char *) "Practice2", (XtActionProc) PracticeBarrel2},
	{(char *) "Randomize", (XtActionProc) RandomizeBarrel},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizeBarrelMaybe},
	{(char *) "Randomize2", (XtActionProc) RandomizeBarrel2},
	{(char *) "Get", (XtActionProc) GetBarrel},
	{(char *) "Write", (XtActionProc) WriteBarrel},
	{(char *) "Clear", (XtActionProc) ClearBarrel},
	{(char *) "Undo", (XtActionProc) UndoBarrel},
	{(char *) "Solve", (XtActionProc) SolveBarrel},
	{(char *) "Orientize", (XtActionProc) OrientizeBarrel},
	{(char *) "Pairs", (XtActionProc) PairsBarrel},
	{(char *) "Enter", (XtActionProc) EnterBarrel},
	{(char *) "Leave", (XtActionProc) LeaveBarrel}
};

static XtResource resourcesBarrel[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.username),
	 XtRString, (caddr_t) "guest"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(BarrelWidget, barrel.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(BarrelWidget, barrel.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(BarrelWidget, barrel.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNtileColor, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(BarrelWidget, barrel.tileColor),
	 XtRString, (caddr_t) "gray75" /*XtDefaultForeground*/},
	{XtNtileBorder, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(BarrelWidget, barrel.borderColor),
	 XtRString, (caddr_t) "gray50" /*XtDefaultBackground*/},
	{XtNfaceColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[0]),
	 XtRString, (caddr_t) "Green"},
	{XtNfaceColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[1]),
	 XtRString, (caddr_t) "Red"},
	{XtNfaceColor2, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[2]),
	 XtRString, (caddr_t) "Yellow"},
	{XtNfaceColor3, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[3]),
	 XtRString, (caddr_t) "Orange"},
	{XtNfaceColor4, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[4]),
	 XtRString, (caddr_t) "Blue"},
	{XtNfaceColor5, XtCLabel, XtRString, sizeof (String),
	 XtOffset(BarrelWidget, barrel.faceName[5]),
	 XtRString, (caddr_t) "black"},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(BarrelWidget, core.width),
	 XtRString, (caddr_t) "200"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(BarrelWidget, core.height),
	 XtRString, (caddr_t) "200"},
	{XtNtiles, XtCTiles, XtRInt, sizeof (int),
	 XtOffset(BarrelWidget, barrel.tiles),
	 XtRString, (caddr_t) "6"}, /* DEFAULTTILES */
	{XtNfaces, XtCFaces, XtRInt, sizeof (int),
	 XtOffset(BarrelWidget, barrel.faces),
	 XtRString, (caddr_t) "5"}, /* DEFAULTFACES */
	{XtNorient, XtCOrient, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.orient),
	 XtRString, (caddr_t) "FALSE"}, /* DEFAULTORIENT */
	{XtNpairs, XtCPairs, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.pairs),
	 XtRString, (caddr_t) "TRUE"}, /* DEFAULTPAIRS */
	{XtNpractice, XtCPractice, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.practice),
	 XtRString, (caddr_t) "FALSE"}, /* DEFAULTPRACTICE */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(BarrelWidget, barrel.delay),
	 XtRString, (caddr_t) "10"}, /* DEFAULTDELAY */
	{XtNbase, XtCBase, XtRInt, sizeof (int),
	 XtOffset(BarrelWidget, barrel.base),
	 XtRString, (caddr_t) "10"}, /* DEFAULTBASE */
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNcheat, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(BarrelWidget, barrel.cheat),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(BarrelWidget, barrel.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(BarrelWidget, barrel.select),
	 XtRCallback, (caddr_t) NULL}
};

BarrelClassRec barrelClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Barrel",	/* class name */
		sizeof (BarrelRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeBarrel,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListBarrel,	/* actions */
		XtNumber(actionsListBarrel),	/* num actions */
		resourcesBarrel,	/* resources */
		XtNumber(resourcesBarrel),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyBarrel,	/* destroy */
		(XtWidgetProc) ResizeBarrel,	/* resize */
		(XtExposeProc) ExposeBarrel,	/* expose */
		(XtSetValuesFunc) SetValuesBarrel,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsBarrel,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass barrelWidgetClass = (WidgetClass) & barrelClassRec;

#ifndef HAVE_USLEEP
#if !defined( VMS ) || defined( XVMSUTILS ) ||  ( __VMS_VER >= 70000000 )
#ifdef USE_XVMSUTILS
#include <X11/unix_time.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#endif
#if defined(SYSV) || defined(SVR4)
#ifdef LESS_THAN_AIX3_2
#include <sys/poll.h>
#else /* !LESS_THAN_AIX3_2 */
#include <poll.h>
#endif /* !LESS_THAN_AIX3_2 */
#endif /* defined(SYSV) || defined(SVR4) */

/* not static in case usleep found in system include */
int
usleep(unsigned int usec)
{
#if (defined (SYSV) || defined(SVR4)) && !defined(__hpux)
#if defined(HAVE_NANOSLEEP)
	{
		struct timespec rqt;

		rqt.tv_nsec = 1000 * (usec % (unsigned int) 1000000);
		rqt.tv_sec = usec / (unsigned int) 1000000;
		return nanosleep(&rqt, NULL);
	}
#else
	(void) poll(
#if defined(__cplusplus) || defined(c_plusplus)
		(pollfd *) /* guess */
#else
		(void *)
#endif
		0, (int) 0, usec / 1000);	/* ms resolution */
#endif
#else
#ifdef VMS
	long        timadr[2];

	if (usec != 0) {
		timadr[0] = -usec * 10;
		timadr[1] = -1;

		sys$setimr(4, &timadr, 0, 0, 0);
		sys$waitfr(4);
	}
#else
	struct timeval time_out;

#if  0
	/* (!defined(AIXV3) && !defined(__hpux)) */
	extern int  select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

#endif

	time_out.tv_usec = usec % (unsigned int) 1000000;
	time_out.tv_sec = usec / (unsigned int) 1000000;
	(void) select(0, (void *) 0, (void *) 0, (void *) 0, &time_out);
#endif
#endif
	return 0;
}
#endif

void
SetBarrel(BarrelWidget w, int reason)
{
	barrelCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#ifndef LOGFILE
#define LOGFILE "barrel.log"
#endif

#if (!defined(WINVER) || (WINVER <= 0x030a)) /* if X or WINDOWS 3.1 or less */
static void
Sleep(unsigned int cMilliseconds)
{
#if (defined(WINVER) && (WINVER <= 0x030a))
	unsigned long time_out = GetTickCount() + cMilliseconds;

	while (time_out > GetTickCount());
#else
	(void) usleep(cMilliseconds * 1000);
#endif
}
#endif

void
intCat(char ** string, const char * var1, const int var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + 21))) {
        	DISPLAY_ERROR("Not enough memory, exiting.");
	}
	(void) sprintf(*string, "%s%d", var1, var2);
}

void
stringCat(char ** string, const char * var1, const char * var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + strlen(var2) + 1))) {
        	DISPLAY_ERROR("Not enough memory, exiting.");
	}
	(void) sprintf(*string, "%s%s", var1, var2);
}

static void
CheckTiles(BarrelWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->barrel.tiles < MINTILES) {
		intCat(&buf1,
			"Number of tiles in X direction out of bounds, use at least ",
			MINTILES);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTTILES);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->barrel.tiles = DEFAULTTILES;
	}
	if (w->barrel.faces < MINFACES || w->barrel.faces > MAXFACES) {
		intCat(&buf1,
			"Number of faces out of bounds, use ",
			MINFACES);
		stringCat(&buf2, buf1, "..");
		free(buf1);
		intCat(&buf1, buf2, MAXFACES);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTFACES);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->barrel.faces = DEFAULTFACES;
	}
	if (w->barrel.base < MINBASE  || w->barrel.base > MAXBASE) {
		intCat(&buf1, "Base out of bounds, use ", MINBASE);
		stringCat(&buf2, buf1, "..");
		free(buf1);
		intCat(&buf1, buf2, MAXBASE);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTBASE);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->barrel.base = DEFAULTBASE;
	}
}

static      Boolean
PositionTile(BarrelWidget w, int x, int *r)
{
	*r = (x - w->barrel.delta.x / 2 - w->barrel.puzzleOffset.x) / w->barrel.offset.x;
	if (*r < 0 || *r >= w->barrel.tiles)
		return False;
	return True;
}

static int
Column(BarrelWidget w, int pos)
{
	return (pos % w->barrel.tiles);
}

static int
Row(BarrelWidget w, int pos)
{
	return (pos / w->barrel.tiles);
}

static int
TileNFromSpace(BarrelWidget w, int plunger, int n, int direction)
{
	int pos = w->barrel.spacePosition[plunger];

	if (direction == LEFT || direction == RIGHT)
		return (pos + ((direction == RIGHT) ? -n : n));
	else /* direction == TOP || direction == BOTTOM */
		return (pos + (w->barrel.tiles * ((direction == BOTTOM) ?
			-n : n)));
}

Boolean
CheckSolved(const BarrelWidget w)
{
	int color, i, j, inc;

	if (w->barrel.tiles < 2)
		return True;
	if (w->barrel.orient) {
		for (j = 0; j < w->barrel.faces; j++) {
			color = (w->barrel.tileOfPosition[j * w->barrel.tiles + 1] - 1) / (w->barrel.tiles - 2);
			inc = (w->barrel.tileOfPosition[j * w->barrel.tiles + 1] - 1) % (w->barrel.tiles - 2);
			for (i = 1; i < w->barrel.tiles - 2; i++) {
				if  (color != (w->barrel.tileOfPosition[j * w->barrel.tiles + i + 1] - 1) / (w->barrel.tiles - 2)) {
					return False;
				}
				if  (inc + i != (w->barrel.tileOfPosition[j * w->barrel.tiles + i + 1] - 1) % (w->barrel.tiles - 2)) {
					return False;
				}
			}
			if (!(color % 2)) {
				if ((w->barrel.tiles - 2) * w->barrel.faces + 1 + color / 2 != w->barrel.tileOfPosition[j * w->barrel.tiles + w->barrel.tiles - 1]) 
					return False;
			}
		}
	} else {

		for (j = 0; j < w->barrel.faces; j++) {
			color = (w->barrel.tileOfPosition[j * w->barrel.tiles + 1] - 1) / (w->barrel.tiles - 2);
			for (i = 0; i < w->barrel.tiles - 3; i++) {
				if  (color != (w->barrel.tileOfPosition[j * w->barrel.tiles + i + 2] - 1) / (w->barrel.tiles - 2)) {
					return False;
				}
			}
		}
	}
	return True;
}

#ifdef DEBUG
static void
PrintBarrel(BarrelWidget w)
{
	int         face, tile, plunger;

	for (face = 0; face < w->barrel.faces; face++) {
		for (tile = 0; tile < w->barrel.tiles; tile++) {
			(void) printf("%d ",
				w->barrel.tileOfPosition[face *
				w->barrel.tiles + tile]);
			if (!((tile + 1) % w->barrel.tiles))
				(void) printf("\n");
		}
	}
	(void) printf("\n");
	for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
		(void) printf("%d: %d\n", plunger,
			w->barrel.spacePosition[plunger]);
	}
	(void) printf("\n");
}
#endif

static int
int2String(BarrelWidget w, char *buf, int number, int base, Boolean capital)
{
	int         digit, mult = base, last, position;
	int         a, i, j, s, r;

	if (capital) {
		a = 'A', i = 'I', j = 'J', s = 'S', r = 'R';
	} else {
		a = 'a', i = 'i', j = 'j', s = 's', r = 'r';
	}
	if (number < 0) {
		char *buff;

		intCat(&buff, "int2String: 0 > ", number);
		DISPLAY_WARNING(buff);
		free(buff);
		return 0;
	}
	last = 1;
	while (number >= mult) {
		last++;
		mult *= base;
	}
	for (position = 0; position < last; position++) {
		mult /= base;
		digit = number / mult;
		number -= digit * mult;
		buf[position] = digit + '0';
		if (buf[position] > '9') {	/* ASCII */
			buf[position] += (a - '9' - 1);
		} else if (buf[position] < '0') {	/* EBCDIC */
			buf[position] += (a - '9' - 1);
			if (buf[position] > i)
				buf[position] += (j - i - 1);
			if (buf[position] > r)
				buf[position] += (s - r - 1);
		}
	}
	buf[last] = '\0';
	return last;
}

static void
fill3DRect(BarrelWidget w, GC gc, GC darkerGC, GC brighterGC,
         int x, int y, int width, int height, Boolean raised)
{
        GC currentGC = (raised) ? gc : darkerGC;

        if (width > 2 && height > 2)
                FILLRECTANGLE(w, currentGC,
                        x + 1, y + 1, width - 2, height - 2);
        currentGC = (raised) ? brighterGC : darkerGC;
        FILLRECTANGLE(w, currentGC,
                x, y, 1, height);
        if (width > 1)
                FILLRECTANGLE(w, currentGC,
                        x + 1, y, width - 2, 1);
        currentGC = (raised) ? darkerGC : gc;
        if (width > 1 && height > 1)
                FILLRECTANGLE(w, currentGC,
                        x, y + height - 1, width, 1);
        if (width > 1 && height > 1)
                FILLRECTANGLE(w, currentGC,
                        x + width - 1, y, 1, height - 1);
}

#define MULT 64
static void
DrawBall(BarrelWidget w, GC gc, int pos, int offsetX, int offsetY)
{
	int         dx, dy, diameter;

	diameter = MIN(w->barrel.offset.x, w->barrel.offset.y) * 3 / 4 - 3;
	dx = Column(w, pos) * w->barrel.offset.x + w->barrel.delta.x +
		w->barrel.puzzleOffset.x + w->barrel.tileSize.x / 2 +
	       	- 1 + offsetX;
	dy = Row(w, pos) * w->barrel.offset.y + w->barrel.delta.y +
		w->barrel.puzzleOffset.y + w->barrel.tileSize.y / 2
	       	- 2 + offsetY;
	FILLCIRCLE(w, gc, diameter, dx, dy);
}

static void
DrawTile(BarrelWidget w, int pos, Boolean blank, Boolean erase,
		int pressedOffset, int offsetX, int offsetY)
{
	int         dx, dy, tile = w->barrel.tileOfPosition[pos];
	GC          ballGC, tileGC, borderGC;

	if (erase) {
		tileGC = w->barrel.inverseGC;
		borderGC = w->barrel.inverseGC;
	} else if (pressedOffset) {
		tileGC = w->barrel.borderGC;
		borderGC = w->barrel.tileGC;
	} else {
		tileGC = w->barrel.tileGC;
		borderGC = w->barrel.borderGC;
	}
	if (w->barrel.mono && pressedOffset) {
		ballGC = borderGC;
	} else {
		int face = ((tile - 1) / (w->barrel.tiles - 2)) % MAXCOLORS;

		ballGC = w->barrel.faceGC[face];
#ifndef WINVER
		if (!w->barrel.mono) {
			if (w->barrel.faceColor[face] == w->barrel.tileColor)  {
				ballGC = (pressedOffset) ? tileGC : borderGC;
			}
		}
#endif
	}
	dx = Column(w, pos) * w->barrel.offset.x + w->barrel.delta.x +
		w->barrel.puzzleOffset.x + pressedOffset - 1;
	dy = Row(w, pos) * w->barrel.offset.y + w->barrel.delta.y +
		w->barrel.puzzleOffset.y + pressedOffset - 1;
	dx += offsetX;
	dy += offsetY;
	if (blank) {
		FILLRECTANGLE(w, tileGC, dx, dy,
		       w->barrel.tileSize.x + 1, w->barrel.tileSize.y - 1);
	} else {
		DRAWRECTANGLE(w, borderGC, dx, dy,
			w->barrel.tileSize.x + 1, w->barrel.tileSize.y - 1);
		fill3DRect(w, tileGC,
			w->barrel.tileDarkerGC, w->barrel.tileBrighterGC,
                        dx + 1, dy + 1,
			w->barrel.tileSize.x, w->barrel.tileSize.y - 2,
                        pressedOffset == 0);
		/* FILLRECTANGLE(w, tileGC,
		       dx + 1, dy + 1,
		       w->barrel.tileSize.x, w->barrel.tileSize.y - 2); */
	}
	if (tile > 0 && tile < ((w->barrel.tiles - 2) * w->barrel.faces +
			NUMPLUNGERS + 1)) {
		DrawBall(w, ballGC, pos, offsetX + pressedOffset,
			offsetY + pressedOffset);
	}
	if (!blank) {
		int         i = 0, digitX = -2, t = tile;
		char        buf[66];

		buf[0] = '\0';
		if (w->barrel.orient || w->barrel.mono) {
			if (w->barrel.mono) {
				if (w->barrel.orient) {
					int         last;

					last = int2String(w, buf, tile, w->barrel.base, False);
					buf[last] =
#ifdef WINVER
						w->barrel.faceChar[((tile - 1) / (w->barrel.tiles - 2)) %
							  MAXCOLORS];
#else
						w->barrel.faceName[((tile - 1) / (w->barrel.tiles -2)) %
							  MAXCOLORS][0];
#endif
					buf[last + 1] = '\0';
					t *= w->barrel.base;
				} else {
					buf[0] =
#ifdef WINVER
						w->barrel.faceChar[((tile - 1) /
							(w->barrel.tiles - 2)) % MAXCOLORS];
#else
						w->barrel.faceName[((tile - 1) /
							(w->barrel.tiles - 2)) % MAXCOLORS][0];
#endif
					buf[1] = '\0';
					t = 1;
				}
			} else {
				(void) int2String(w, buf, tile, w->barrel.base, True);
			}
			while (t >= 1) {
				t /= w->barrel.base;
				digitX += w->barrel.digitOffset.x;
				i++;
			}
			DRAWTEXT(w, tileGC,
				    dx + w->barrel.tileSize.x / 2 - digitX,
				    dy + w->barrel.tileSize.y / 2 + w->barrel.digitOffset.y,
				    buf, i);
		}
	}
#ifndef WINVER
	DRAWRECTANGLE(w, borderGC, dx, dy,
		       w->barrel.tileSize.x + 1, w->barrel.tileSize.y - 1);
#endif
}

void
DrawAllTiles(const BarrelWidget w)
{
	int         k;

	for (k = 0; k < w->barrel.tileFaces; k++)
		DrawTile(w, k, (w->barrel.tileOfPosition[k] < 0),
			 (w->barrel.tileOfPosition[k] < 0), FALSE, 0, 0);
}

static void
AnimateSlide(BarrelWidget w, int numTiles, int dir, int fast)
{
	int inc, aTile, plunger;
	int fillBeginPos[NUMPLUNGERS], fillBeginI = 0, gapI, moveI;
	int dx, dy, pos, posNext;
	int ix = 0, iy = 0;

	for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
	  fillBeginPos[plunger] = TileNFromSpace(w, plunger, numTiles, dir);
	}
	fillBeginI = Column(w, fillBeginPos[0]) * w->barrel.offset.x +
	  w->barrel.delta.x + w->barrel.puzzleOffset.x;
	gapI = w->barrel.tileSize.x * fast / w->barrel.numSlices;
	moveI = w->barrel.tileSize.x + w->barrel.delta.x; 
	if (gapI == 0)
	  gapI++;
	if (numTiles < 0)
	  numTiles = -numTiles;
	for (inc = 0; inc < moveI + gapI; inc += gapI) {
	  if (inc > moveI)
	    inc = moveI;
	  for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
	    for (aTile = 0; aTile < numTiles; aTile++) {
	      posNext = TileNFromSpace(w, plunger, aTile + 1, dir);
	      /* Calculate deltas */
	      dx = Column(w, posNext) * w->barrel.offset.x + w->barrel.delta.x +
	        w->barrel.puzzleOffset.x;
	      dy = Row(w, posNext) * w->barrel.offset.y + w->barrel.delta.y +
	        w->barrel.puzzleOffset.y;
	      ix = ((dir == RIGHT) ? inc : -inc);
	      iy = 0;
	      DrawTile(w, posNext, False, False, FALSE, ix, iy);
	      /* Erase old slivers */
	      ix += dx;
	      iy += dy;
	      if (aTile < numTiles - 1) {
	        switch (dir) {
	          case RIGHT:
	            if (gapI - 1 > 0)
	              FILLRECTANGLE(w,
	                w->barrel.inverseGC,
	                ix - gapI, iy - 1, gapI - 1, w->barrel.tileSize.y + 1);
	              break;
	          case LEFT:
	            if (gapI - 1 > 0)
	              FILLRECTANGLE(w,
	                w->barrel.inverseGC,
	                ix + w->barrel.tileSize.x + 1, iy - 1,
	                gapI - 1, w->barrel.tileSize.y + 1);
		    break;
	        }
	      } else {
	        switch (dir) {
	          case RIGHT:
	            if (ix - fillBeginI - 1 > 0)
	              FILLRECTANGLE(w,
	                w->barrel.inverseGC,
	                fillBeginI - 1, iy - 1,
	                ix - fillBeginI, w->barrel.tileSize.y + 1);
	            break;
	          case LEFT:
	            FILLRECTANGLE(w,
	              w->barrel.inverseGC,
	              ix + w->barrel.tileSize.x + 1, iy - 1,
	              fillBeginI - ix, w->barrel.tileSize.y + 1);
		    break;
	        }
	      }
	    }
#ifndef WINVER
	    XFlush(XtDisplay(w));
#endif
	    Sleep((unsigned int) (w->barrel.delay / fast));
	  }
	}
	for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
	  pos = w->barrel.spacePosition[plunger];
	  for (aTile = 0; aTile < numTiles; aTile++) {
	    posNext = TileNFromSpace(w, plunger, aTile + 1, dir);
	    w->barrel.tileOfPosition[pos] =
	      w->barrel.tileOfPosition[posNext];
	    pos = posNext;
	  }
	  w->barrel.spacePosition[plunger] = fillBeginPos[plunger];
	  w->barrel.tileOfPosition[fillBeginPos[plunger]] = -NUMPLUNGERS +
		  plunger;
	}
}

static void
DrawAllTilesForColumns(BarrelWidget w, int start, int finish)
{
	int         i, j, k;

	for (i = start; i <= finish; i++) {
		for (j = 0; j < w->barrel.faces; j++) {
			k = j * w->barrel.tiles + i;
			if (w->barrel.tileOfPosition[k] > 0)
			  DrawTile(w, k,
			    (w->barrel.tileOfPosition[k] < 0),
			    (w->barrel.tileOfPosition[k] < 0), FALSE, 0, 0);
		}
	}
}

static void
AnimateRotate(BarrelWidget w, int start, int finish, int direction, int fast)
{
#ifdef WINVER
	HDC memDC;
	HBITMAP hBit;
#else
	Pixmap bufferBox;
	XWindowAttributes xgwa;
#endif
	int size = finish - start + 1;
	int moveI, gapJ, moveJ, inc, i;

	moveI = w->barrel.tileSize.x + w->barrel.delta.x;
	gapJ = w->barrel.tileSize.y * fast / w->barrel.numSlices;
	moveJ = w->barrel.tileSize.y + w->barrel.delta.y + 1;
	if (gapJ == 0)
		gapJ++;
	
#ifdef WINVER
	memDC = CreateCompatibleDC(w->core.hDC);
	hBit = CreateCompatibleBitmap(w->core.hDC, size * moveI + 1,
		w->barrel.puzzleSize.y);
	SelectObject(memDC, hBit);
	BitBlt(memDC, 0, 0,
		size * moveI + 1,
		w->barrel.puzzleSize.y,
		w->core.hDC,
		start * moveI + w->barrel.puzzleOffset.x + 1,
		w->barrel.puzzleOffset.y,
		SRCCOPY);
#else
	(void) XGetWindowAttributes(XtDisplay(w), XtWindow(w), &xgwa);

	if ((bufferBox = XCreatePixmap(XtDisplay(w), XtWindow(w),
       			size * moveI + 1,
			w->barrel.puzzleSize.y,
			xgwa.depth)) == None) {
		XtError("Not enough memory, exiting.");
	}
	XCopyArea(XtDisplay(w), XtWindow(w), bufferBox,
		w->barrel.frameGC,
		start * moveI + w->barrel.puzzleOffset.x + 1,
		w->barrel.puzzleOffset.y,
		size * moveI + 1,
		w->barrel.puzzleSize.y,
		0, 0);
	XSetGraphicsExposures(XtDisplay(w), w->barrel.frameGC, False);
#endif
	for (inc = 0; inc < moveJ + gapJ; inc += gapJ) {
		if (inc > moveJ)
			inc = moveJ;
		if (direction == TOP)
			i = inc;
		else
			i = w->barrel.puzzleSize.y - inc -
				w->barrel.puzzleOffset.y + 1;
#ifdef WINVER
		SelectObject(memDC, hBit);
		BitBlt(w->core.hDC,
			start * moveI + w->barrel.puzzleOffset.x + 1,
			w->barrel.puzzleOffset.y + 1,
			size * moveI + 1,
			w->barrel.puzzleSize.y - i,
			memDC,
			0, i,
			SRCCOPY);
		BitBlt(w->core.hDC,
			start * moveI + w->barrel.puzzleOffset.x + 1,
			w->barrel.puzzleSize.y - i + w->barrel.puzzleOffset.y -
			w->barrel.delta.y + 1,
			size * moveI + 1,
			i - w->barrel.delta.y + w->barrel.puzzleOffset.y + 1,
			memDC,
			0, 0,
			SRCCOPY);
#else

		XCopyArea(XtDisplay(w), bufferBox, XtWindow(w),
			w->barrel.frameGC,
			0, i,
			size * moveI + 1,
			w->barrel.puzzleSize.y - i,
			start * moveI + w->barrel.puzzleOffset.x + 1,
			w->barrel.puzzleOffset.y + 1);
		XCopyArea(XtDisplay(w), bufferBox, XtWindow(w),
			w->barrel.frameGC,
			0, 0,
			size * moveI + 1,
			i - w->barrel.delta.y + w->barrel.puzzleOffset.y + 1,
			start * moveI + w->barrel.puzzleOffset.x + 1,
			w->barrel.puzzleSize.y - i + w->barrel.puzzleOffset.y -
			w->barrel.delta.y + 1);
		XFlush(XtDisplay(w));
#endif
		Sleep((unsigned int) (w->barrel.delay / fast));
	}
#ifdef WINVER
	DeleteObject(hBit);
	DeleteDC(memDC);
#else
	XFreePixmap(XtDisplay(w), bufferBox);
#endif
}

static void
ResizeTiles(BarrelWidget w)
{
	w->barrel.digitOffset.x = 3;
	w->barrel.digitOffset.y = 4;
}

static void
ResetTiles(BarrelWidget w)
{
	int         i, inc;

	w->barrel.tileFaces = w->barrel.tiles * w->barrel.faces;
	if (w->barrel.tileOfPosition)
		free(w->barrel.tileOfPosition);
	if (!(w->barrel.tileOfPosition = (int *)
	      malloc(sizeof (int) * w->barrel.tileFaces))) {
		DISPLAY_ERROR("Not enough memory, exiting.");
	}
	if (startPosition)
		free(startPosition);
	if (!(startPosition = (int *)
	      malloc(sizeof (int) * w->barrel.tileFaces))) {
		DISPLAY_ERROR("Not enough memory, exiting.");
	}

	inc = 0;
	for (i = 0; i < w->barrel.tileFaces; i++) {
		if (i % w->barrel.tiles != 0 &&
				i % w->barrel.tiles != w->barrel.tiles - 1) {
			inc++;
			w->barrel.tileOfPosition[i] = inc;
		} else {
			w->barrel.tileOfPosition[i] = 0;
		}
	}
	for (i = 0; i < NUMPLUNGERS; i++) {
		w->barrel.spacePosition[i] = 2 * i * w->barrel.tiles;
		w->barrel.tileOfPosition[2 * i * w->barrel.tiles] =
			-NUMPLUNGERS + i;
		inc++;
		w->barrel.tileOfPosition[(2 * i + 1) * w->barrel.tiles - 1] =
			inc;
	}
	inc++;
	for (i = 0; i < w->barrel.faces - NUMPLUNGERS; i++) {
		w->barrel.tileOfPosition[(2 * i + 1) * w->barrel.tiles] =
			inc;
		w->barrel.tileOfPosition[(2 * i + 2) * w->barrel.tiles - 1] =
			inc;
	}
	FlushMoves(w);
	w->barrel.currentRef = -w->barrel.tileFaces;
	w->barrel.started = False;
}

static void
EraseFrame(const BarrelWidget w)
{
	FILLRECTANGLE(w, w->barrel.inverseGC,
		0, 0, w->core.width, w->core.height);
}

static void
DrawFrame(BarrelWidget w, Boolean focus)
{
	int         sumX, sumY, offsetX, offsetY, r;
	GC gc = (focus) ? w->barrel.frameGC : w->barrel.borderGC;

	offsetX = w->barrel.puzzleOffset.x;
	offsetY = w->barrel.puzzleOffset.y;
	sumX = w->barrel.tiles * w->barrel.offset.x + w->barrel.delta.x / 2 + 1;
	sumY = w->barrel.offset.y;
	FILLRECTANGLE(w, w->barrel.borderGC,
		0, 0, w->core.width, offsetY);
	FILLRECTANGLE(w, w->barrel.borderGC,
		0, w->core.height - offsetY - 1,
		w->core.width, offsetY + 2);
	FILLRECTANGLE(w, w->barrel.borderGC,
		0, 0, offsetX, w->core.height);
	FILLRECTANGLE(w, w->barrel.borderGC,
		offsetX + sumX, 0,
		w->core.width - sumX - offsetX, w->core.height);
	FILLRECTANGLE(w, gc,
		offsetX, offsetY, sumX, 1);
	FILLRECTANGLE(w, gc,
		offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	FILLRECTANGLE(w, gc,
		sumX + offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	for (r = 1; r < w->barrel.faces; r++) {
		offsetY += sumY;
		FILLRECTANGLE(w, gc,
			       offsetX, offsetY, sumX, 1);
	}
	FILLRECTANGLE(w, gc,
	       offsetX, offsetY + sumY, sumX + 1, 1);
}

static void
MoveNoTiles(BarrelWidget w)
{
	SetBarrel(w, BARREL_IGNORE);
}

static void
MoveTiles(BarrelWidget w, int from, int fast)
{
	w->barrel.currentRef = from;
	if (fast != INSTANT && w->barrel.delay > 0) {
		if (w->barrel.spacePosition[0] % w->barrel.tiles == 0) {
			AnimateSlide(w, w->barrel.tiles - 1, LEFT, fast);
		} else {
			AnimateSlide(w, w->barrel.tiles - 1, RIGHT, fast);
		}
	} else {
		int  pos, posNext, plunger, aTile, dir;

		dir = (w->barrel.spacePosition[0] % w->barrel.tiles == 0) ?
			LEFT : RIGHT;
		for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
			pos = w->barrel.spacePosition[plunger];
			for (aTile = 0; aTile < w->barrel.tiles - 1; aTile++) {
				posNext = TileNFromSpace(w, plunger, 1, dir);
				w->barrel.tileOfPosition[pos] =
					w->barrel.tileOfPosition[posNext];
				DrawTile(w, pos, False, False, FALSE, 0, 0);
				pos = posNext;
				w->barrel.spacePosition[plunger] = pos;
			}
	  		w->barrel.tileOfPosition[pos] = -NUMPLUNGERS + plunger;
			DrawTile(w, pos, True, True, FALSE, 0, 0);
		}
	}
	w->barrel.currentRef = -w->barrel.tileFaces;
}

static void
RotateTiles(BarrelWidget w, int start, int finish, int direction, int fast)
{
	int         newR, i, k, tempTile, currTile, pos, newPos;
	int         r = 0, plunger;
	Boolean	    animate = (fast != INSTANT && w->barrel.delay > 0);

	if (animate)
		AnimateRotate(w, start, finish, direction, fast);
	for  (i = start; i <= finish; i++) {
		currTile = w->barrel.tileOfPosition[r * w->barrel.tiles + i];
		for (k = 1; k <= w->barrel.faces; k++) {
			newR = (direction == TOP) ?  (r +
				w->barrel.faces - 1) % w->barrel.faces :
				(r + 1) % w->barrel.faces;
			pos = r * w->barrel.tiles + i;
			newPos = newR * w->barrel.tiles + i;
			tempTile = w->barrel.tileOfPosition[newPos];
			w->barrel.tileOfPosition[newPos] = currTile;
			currTile = tempTile;
			for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
				if (w->barrel.spacePosition[plunger] == pos) {
					if (!animate)
						DrawTile(w, newPos, True, True,
							FALSE, 0, 0);
					break;
				}
			}
			if (!animate && plunger == NUMPLUNGERS)
				DrawTile(w, newPos, False, False,
					FALSE, 0, 0);
			r = newR;
		}
	}
	if (start == 0)  {
		for (plunger = 0; plunger < NUMPLUNGERS; plunger++) {
			i = (direction == TOP) ? w->barrel.faces - 1 : 1;
			w->barrel.spacePosition[plunger] =
				(w->barrel.spacePosition[plunger] + i *
					w->barrel.tiles) % w->barrel.tileFaces;
		}
	}
	/* Would be nice if this was not needed. */
	if (animate) {
		int moveI = w->barrel.tileSize.x + w->barrel.delta.x;
		
		FILLRECTANGLE(w, w->barrel.inverseGC,
			start * moveI + w->barrel.puzzleOffset.x + 1, 0,
			(finish - start + 1) * moveI + 1, w->core.height);
		DrawFrame(w, w->barrel.focus);
		DrawAllTilesForColumns(w, start, finish);
	}
}

static void
SelectSlideTiles(BarrelWidget w, int pos)
{
	if (pos < 0) {
		if (w->barrel.delay > 0)
			AnimateSlide(w, -pos, RIGHT, NORMAL);
		else {
			MoveTiles(w, w->barrel.spacePosition[0] % w->barrel.tiles - w->barrel.tiles + 1,
				NORMAL);
		}
		SetBarrel(w, BARREL_MOVED);
		PutMove(RIGHT, 0, 0);
	} else if (pos > 0) {
		if (w->barrel.delay > 0)
			AnimateSlide(w, pos, LEFT, NORMAL);
		else {
			MoveTiles(w, w->barrel.spacePosition[0] % w->barrel.tiles + w->barrel.tiles - 1,
				NORMAL);
		}
		SetBarrel(w, BARREL_MOVED);
		PutMove(LEFT, w->barrel.tiles - 1, 0);
	}
}

static      Boolean
MoveTilesDir(BarrelWidget w, int direction, int tile, Boolean all, int fast)
{
	int start, finish;

	start = tile;
	finish = tile;
	if (all) {
		start = 0;
		finish = w->barrel.tiles - 1;
	} else if (w->barrel.pairs && tile > 0 && tile < w->barrel.tiles - 1) {
		start = 1 + ((tile - 1) / 2) * 2;
		finish =  start + 1;
	}
	switch (direction) {
	case TOP:
		RotateTiles(w, start, finish, TOP, fast);
		return True;
	case RIGHT:
		if (Column(w, w->barrel.spacePosition[0]) > 0) {
			MoveTiles(w, 0, fast);
			return True;
		}
		break;
	case BOTTOM:
		RotateTiles(w, start, finish, BOTTOM, fast);
		return True;
	case LEFT:
		if (Column(w, w->barrel.spacePosition[0]) < w->barrel.tiles - 1) {
			MoveTiles(w, w->barrel.tiles - 1, fast);
			return True;
		}
		break;
	default:
		{
			char *buf;

			intCat(&buf, "MoveTilesDir: direction ", direction);
			DISPLAY_WARNING(buf);
			free(buf);
		}
	}
	return False;
}

#if 0
static void
ControlBarrel(BarrelWidget w, int direction)
{
	int reason = BARREL_IGNORE;

	switch (direction) {
		case TOP:
			if (w->barrel.faces <= MINFACES)
				return;
			reason = BARREL_DEC_Y;
			break;
		case RIGHT:
			reason = BARREL_INC_X;
			break;
		case BOTTOM:
			if (w->barrel.faces >= MAXFACES)
				return;
			reason = BARREL_INC_Y;
			break;
		case LEFT:
			if (w->barrel.tiles <= MINTILES)
				return;
			reason = BARREL_DEC_X;
			break;
		default:
			{
				char *buf;

				intCat(&buf, "ControlBarrel: direction ",
					direction);
				DISPLAY_WARNING(buf);
				free(buf);
			}
	}
	SetBarrel(w, reason);
}
#endif

static void
MoveShift(BarrelWidget w, int direction, int fast)
{
	(void) MoveTilesDir(w, direction, 0, True, fast);
	SetBarrel(w, BARREL_CONTROL);
}

static void
SlideBarrel(BarrelWidget w, int direction)
{
	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	}
	if (!MoveBarrel(w, direction, 0, FALSE, NORMAL)) {
		SetBarrel(w, BARREL_BLOCKED);
		return;
	}
	if (CheckSolved(w)) {
		SetBarrel(w, BARREL_SOLVED);
	}
	return;
}

static void
RotateBarrel(BarrelWidget w, int direction, int tile)
{
	if (CheckSolved(w) && !w->barrel.practice) {
		MoveNoTiles(w);
		return;
	}
	(void) MoveBarrel(w, direction, tile, FALSE, NORMAL);
	if (CheckSolved(w)) {
		SetBarrel(w, BARREL_SOLVED);
	}
	return;
}

#ifndef WINVER
static
#endif
void
MoveBarrelInput(BarrelWidget w, int x, int direction, int shift, int control)
{
	int         r;

#if 0
	if (control)
		ControlBarrel(w, direction);
	else
#endif
	if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShift(w, direction, NORMAL);
		PutMove(direction, 0, 1);
	} else if (CheckSolved(w) && !w->barrel.practice) {
		MoveNoTiles(w);
		return;
	} else if (direction == LEFT || direction == RIGHT) {
		SlideBarrel(w, direction);
		if (CheckSolved(w)) {
			SetBarrel(w, BARREL_SOLVED);
		}
	} else {
		if (!PositionTile(w, x, &r))
			return;
		if (r == 0 || r == w->barrel.tiles - 1) {
			SetBarrel(w, BARREL_BLOCKED);
			return;
		}
		RotateBarrel(w, direction, r);
		if (CheckSolved(w)) {
			SetBarrel(w, BARREL_SOLVED);
		}
	}
}

void
MoveBarrelDelay(BarrelWidget w, const int direction, const int tile,
		const Boolean all)
{
	if (direction == RIGHT) {
		PutMove(direction, 0, False);
	} else if (direction == LEFT) {
		PutMove(direction, w->barrel.tiles - 1, False);
	} else {
		SetBarrel(w, BARREL_MOVED);
		PutMove(direction, tile, all);
	}
	(void) MoveTilesDir(w, direction, tile, all, NORMAL);
	Sleep((unsigned int) w->barrel.delay);
}

Boolean
MoveBarrel(BarrelWidget w, const int direction, const int tile,
		const int shift, const int fast)
{
	if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShift(w, direction, fast);
		PutMove(direction, tile, 1);
		return True;
	} else if (MoveTilesDir(w, direction, tile, False, fast)) {
		SetBarrel(w, BARREL_MOVED);
		PutMove(direction, tile, 0);
		return True;
	}
	return False;
}

static int
SelectTiles(BarrelWidget w, int x, int y, int *i, int *j)
{
	*i = (x - w->barrel.delta.x / 2 - w->barrel.puzzleOffset.x) /
		w->barrel.offset.x;
	*j = ((y - w->barrel.delta.y / 2 - w->barrel.puzzleOffset.y) %
	      (w->barrel.faces * w->barrel.offset.y + w->barrel.delta.y - 1)) /
		w->barrel.offset.y;
	if (*i >= 0 && *j >= 0 &&
	    *i < w->barrel.tiles && *j < w->barrel.faces)
		return (*i + w->barrel.tiles * *j -
			w->barrel.spacePosition[0] % w->barrel.tileFaces);
	return -w->barrel.tileFaces;
}

static void
GetTiles(BarrelWidget w)
{
	FILE       *fp;
	int         c, pairs, orient, practice, moves;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "r")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "r")) == NULL) {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
#endif
	}
	FlushMoves(w);
#if 0
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &tiles);
	if (tiles >= MINTILES) {
		for (i = w->barrel.tiles; i < tiles; i++) {
			SetBarrel(w, BARREL_INC_X);
		}
		for (i = w->barrel.tiles; i > tiles; i--) {
			SetBarrel(w, BARREL_DEC_X);
		}
	} else {
		stringCat(&buf1, name, " corrupted: tiles ");
		intCat(&buf2, buf1, tiles);
		free(buf1);
		stringCat(&buf1, buf2, " should be between ");
		free(buf2);
		intCat(&buf2, buf1, MINTILES);
		free(buf1);
		stringCat(&buf1, buf2, " and MAXINT");
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &faces);
	if (faces >= MINFACES && faces <= MAXFACES) {
		for (i = w->barrel.faces; i < faces; i++) {
			SetBarrel(w, BARREL_INC_Y);
		}
		for (i = w->barrel.faces; i > faces; i--) {
			SetBarrel(w, BARREL_DEC_Y);
		}
	} else {
		stringCat(&buf1, name, " corrupted: faces ");
		intCat(&buf2, buf1, faces);
		free(buf1);
		stringCat(&buf1, buf2, " should be between ");
		free(buf2);
		intCat(&buf2, buf1, MINFACES);
		free(buf1);
		stringCat(&buf1, buf2, " and ");
		free(buf2);
		intCat(&buf2, buf1, MAXFACES);
		free(buf1);
		DISPLAY_WARNING(buf2);
		free(buf2);
	}
#endif
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &pairs);
	if (w->barrel.pairs != (Boolean) pairs) {
		SetBarrel(w, BARREL_PAIRS);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &orient);
	if (w->barrel.orient != (Boolean) orient) {
		SetBarrel(w, BARREL_ORIENT);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &practice);
	if (w->barrel.orient != (Boolean) practice) {
		SetBarrel(w, BARREL_PRACTICE);
	}
#ifdef WINVER
	ResetTiles(w);
#endif
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &moves);
	ScanStartPosition(fp, w);
	SetBarrel(w, BARREL_RESTORE);
	SetStartPosition(w);
	ScanMoves(fp, w, moves);
	(void) fclose(fp);
	(void) printf("%s: pairs %d, orient %d, practice %d, moves %d.\n",
		      name, pairs, orient, practice, moves);
	free(lname);
	free(fname);
	w->barrel.cheat = True; /* Assume the worst. */
}

static void
WriteTiles(BarrelWidget w)
{
	FILE       *fp;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "w")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
#endif
	}
#if 0
	(void) fprintf(fp, "tiles%c %d\n", SYMBOL, w->barrel.tiles);
	(void) fprintf(fp, "faces%c %d\n", SYMBOL, w->barrel.faces);
#endif
	(void) fprintf(fp, "pairs%c %d\n", SYMBOL,
		(w->barrel.pairs) ? 1 : 0);
	(void) fprintf(fp, "orient%c %d\n", SYMBOL,
		(w->barrel.orient) ? 1 : 0);
	(void) fprintf(fp, "practice%c %d\n", SYMBOL,
		(w->barrel.practice) ? 1 : 0);
	(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
	PrintStartPosition(fp, w);
	PrintMoves(fp);
	(void) fclose(fp);
	(void) printf("Saved to %s.\n", name);
	free(lname);
	free(fname);
}

static void
ClearTiles(BarrelWidget w)
{
	if (w->barrel.currentRef != -w->barrel.tileFaces)
		return;
	ResetTiles(w);
	DrawAllTiles(w);
	SetBarrel(w, BARREL_RESET);
}

static void
UndoTiles(BarrelWidget w)
{
	if (MadeMoves() && w->barrel.currentRef == -w->barrel.tileFaces) {
		int         direction, tile, shift;

		GetMove(&direction, &tile, &shift);
		direction = (direction + COORD / 2) % COORD;
		if (shift && (direction == TOP || direction == BOTTOM)) {
			MoveShift(w, direction, DOUBLE);
		} else if (MoveTilesDir(w, direction, tile, False, DOUBLE)) {
			SetBarrel(w, BARREL_UNDO);
		} else {
			char *buf1, *buf2;

			intCat(&buf1, "Move ", direction);
			stringCat(&buf2, buf1, " can not be made");
			free(buf1);
			DISPLAY_WARNING(buf2);
			free(buf2);
		}
	}
}

static void
PracticeTiles(BarrelWidget w)
{
	SetBarrel(w, BARREL_PRACTICE);
}

static void
RandomizeTiles(BarrelWidget w)
{
	if (w->barrel.currentRef != -w->barrel.tileFaces)
		return;
	if (w->barrel.practice)
		PracticeTiles(w);
	SetBarrel(w, BARREL_RESET);
	w->barrel.cheat = False;
	/* Now move the space around randomly */
	if (w->barrel.tiles > 1 || w->barrel.faces > 1) {
		int         big = w->barrel.tileFaces + NRAND(4);
		int         lastDirection = -1;
		int         randomDirection;

		if (big > 100)
			big = 100;
#ifdef DEBUG
		big = 3;
#endif
		while (big--) {
			randomDirection = NRAND(COORD);
			if (randomDirection & 1) {
				if (w->barrel.spacePosition[0] % w->barrel.tiles == 0)
					randomDirection = LEFT;
				else
					randomDirection = RIGHT;
			}
			if  ((randomDirection + COORD / 2) % COORD != lastDirection) {
				if (randomDirection & 1) {
				
					(void) MoveBarrel(w, randomDirection, w->barrel.tiles - 1, FALSE, INSTANT);
				} else {
					(void) MoveBarrel(w, randomDirection, NRAND(w->barrel.tiles - 2) + 1, FALSE, INSTANT);
				}
#ifdef DEBUG
				sleep(1);
#endif
			} else {
				big++;
			}
		}
		FlushMoves(w);
		SetBarrel(w, BARREL_RANDOMIZE);
	}
	if (CheckSolved(w)) {
		SetBarrel(w, BARREL_SOLVED);
	}
}

static void
SolveTiles(BarrelWidget w)
{
	if (CheckSolved(w) || w->barrel.currentRef != -w->barrel.tileFaces)
		return;
	/* if (w->barrel.tiles == 6 && w->barrel.faces == 5) */
#ifdef AUTOSOLVE
	SolveSomeTiles(w);
#else
	SetBarrel(w, BARREL_SOLVE_MESSAGE);
#endif
}

static void
OrientizeTiles(BarrelWidget w)
{
	SetBarrel(w, BARREL_ORIENT);
}

static void
PairsTiles(BarrelWidget w)
{
	SetBarrel(w, BARREL_PAIRS);
}

#define FACTOR 0.7
#ifdef WINVER
#define MAXINTENSITY 0xFF
static int
brighter(const int light)
{
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);
	int temp = light;

	if (temp < i)
		temp = i;
	return MIN(temp / FACTOR, MAXINTENSITY);
}

static int
darker(const int light)
{
	return (int) (light * FACTOR);
}

static void
SetValuesBarrel(BarrelWidget w)
{
	struct tagColor {
		int         red, green, blue;
	} color;
	char        szBuf[80], buf[20], charbuf[2];
	int         face;

	w->barrel.tiles = GetPrivateProfileInt(SECTION, "tiles",
		DEFAULTTILES, INIFILE);
	w->barrel.faces = GetPrivateProfileInt(SECTION, "faces",
		DEFAULTFACES, INIFILE);
	w->barrel.orient = (BOOL) GetPrivateProfileInt(SECTION, "orient",
		FALSE, INIFILE);
	w->barrel.pairs = (BOOL) GetPrivateProfileInt(SECTION, "pairs",
		TRUE, INIFILE);
	w->barrel.base = GetPrivateProfileInt(SECTION, "base",
		DEFAULTBASE, INIFILE);
	w->barrel.delay = GetPrivateProfileInt(SECTION, "delay",
		DEFAULTDELAY, INIFILE);
	w->barrel.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->barrel.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->barrel.frameGC = RGB(color.red, color.green, color.blue);
	/* gray75 */
	(void) GetPrivateProfileString(SECTION, "tileColor", "191 191 191",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->barrel.tileGC = RGB(color.red, color.green, color.blue);
	/* gray50 */
	(void) GetPrivateProfileString(SECTION, "tileBorder", "127 127 127",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->barrel.tileBrighterGC = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->barrel.tileDarkerGC = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->barrel.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "174 178 195",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->barrel.inverseGC = RGB(color.red, color.green, color.blue);
	for (face = 0; face < MAXCOLORS; face++) {
		(void) sprintf(buf, "faceColor%d", face);
		(void) GetPrivateProfileString(SECTION,
			buf, faceColorString[face],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->barrel.faceGC[face] =
			RGB(color.red, color.green, color.blue);
		(void) sprintf(buf, "faceChar%d", face);
		charbuf[0] = faceColorChar[face];
		charbuf[1] = '\0';
		(void) GetPrivateProfileString(SECTION, buf, charbuf,
			szBuf, sizeof (szBuf), INIFILE);
		w->barrel.faceChar[face] = szBuf[0];
	}
	(void) GetPrivateProfileString(SECTION, "name", "Guest",
			szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->barrel.username, szBuf);
	w->barrel.username[80] = 0;
}

void
DestroyBarrel(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
#define MAXINTENSITY 0xFFFF
static Pixel
brighter(BarrelWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(BarrelWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
GetColor(BarrelWidget w, int face, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	XColor      colorCell, rgb;

	valueMask = GCForeground | GCBackground;
	if (w->barrel.reverse) {
		values.background = w->barrel.foreground;
	} else {
		values.background = w->barrel.background;
	}
	if (!w->barrel.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				  DefaultColormapOfScreen(XtScreen(w)),
				w->barrel.faceName[face], &colorCell, &rgb)) {
			values.foreground = w->barrel.faceColor[face] = colorCell.pixel;
			if (!init)
				XtReleaseGC((Widget) w, w->barrel.faceGC[face]);
			w->barrel.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
			return;
		} else {
			char *buf1, *buf2;
 
			stringCat(&buf1, "Color name \"",
				w->barrel.faceName[face]);
			stringCat(&buf2, buf1, "\" is not defined for face ");
			free(buf1);
			intCat(&buf1, buf2, face);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (w->barrel.reverse) {
		values.background = w->barrel.background;
		values.foreground = w->barrel.foreground;
	} else {
		values.background = w->barrel.foreground;
		values.foreground = w->barrel.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.faceGC[face]);
	w->barrel.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
}

static void
SetAllColors(BarrelWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	int         face;

	valueMask = GCForeground | GCBackground;

	if (w->barrel.reverse) {
		values.foreground = w->barrel.foreground;
		values.background = w->barrel.background;
	} else {
		values.foreground = w->barrel.background;
		values.background = w->barrel.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.inverseGC);
	w->barrel.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->barrel.mono) {
		if (w->barrel.reverse) {
			values.foreground = w->barrel.background;
			values.background = w->barrel.foreground;
		} else {
			values.foreground = w->barrel.foreground;
			values.background = w->barrel.background;
		}
	} else {
		values.foreground = w->barrel.frameColor;
		values.background = w->barrel.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.frameGC);
	w->barrel.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->barrel.mono) {
		if (w->barrel.reverse) {
			values.background = w->barrel.foreground;
			values.foreground = w->barrel.background;
		} else {
			values.foreground = w->barrel.foreground;
			values.background = w->barrel.background;
		}
	} else {
		values.foreground = w->barrel.tileColor;
		values.background = w->barrel.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.tileGC);
	w->barrel.tileGC = XtGetGC((Widget) w, valueMask, &values);
	if (!w->barrel.mono) {
		values.foreground = darker(w, w->barrel.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.tileBrighterGC);
	w->barrel.tileBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (!w->barrel.mono) {
		values.foreground = brighter(w, w->barrel.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.tileDarkerGC);
	w->barrel.tileDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->barrel.mono) {
		if (w->barrel.reverse) {
			values.foreground = w->barrel.foreground;
			values.background = w->barrel.background;
		} else {
			values.background = w->barrel.foreground;
			values.foreground = w->barrel.background;
		}
	} else {
		values.foreground = w->barrel.borderColor;
		values.background = w->barrel.tileColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->barrel.borderGC);
	w->barrel.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (face = 0; face < MAXCOLORS; face++)
		GetColor(w, face, init);
}

static      Boolean
SetValuesBarrel(Widget current, Widget request, Widget renew)
{
	BarrelWidget c = (BarrelWidget) current, w = (BarrelWidget) renew;
	Boolean     redraw = False, setColors = False;
	Boolean     redrawTiles = False;
	int         face;

	CheckTiles(w);
	for (face = 0; face < MAXFACES; face++) {
		if (strcmp(w->barrel.faceName[face], c->barrel.faceName[face])) {
			setColors = True;
			break;
		}
	}
	if (w->barrel.background != c->barrel.background ||
	    w->barrel.foreground != c->barrel.foreground ||
	    w->barrel.borderColor != c->barrel.borderColor ||
	    w->barrel.tileColor != c->barrel.tileColor ||
	    w->barrel.reverse != c->barrel.reverse ||
	    w->barrel.mono != c->barrel.mono ||
	    setColors) {
		SetAllColors(w, False);
		redrawTiles = True;
	}
	if (w->barrel.tiles != c->barrel.tiles ||
	    w->barrel.faces != c->barrel.faces ||
	    w->barrel.orient != c->barrel.orient ||
	    w->barrel.pairs != c->barrel.pairs ||
	    w->barrel.practice != c->barrel.practice ||
	    w->barrel.base != c->barrel.base) {
		ResetTiles(w);
		ResizeBarrel(w);
		redraw = True;
	} else if (w->barrel.offset.x != c->barrel.offset.x ||
		   w->barrel.offset.y != c->barrel.offset.y) {
		ResizeBarrel(w);
		redraw = True;
	}
	if (w->barrel.delay != c->barrel.delay) {
		w->barrel.numSlices = ((w->barrel.delay < MAXSLICES) ?
			w->barrel.delay + 1 : MAXSLICES);
	}
	if (w->barrel.menu != -1) {
		switch (w->barrel.menu) {
		case 0:
			w->barrel.menu = -1;
			GetTiles(w);
			break;
		case 1:
			w->barrel.menu = -1;
			WriteTiles(w);
			break;
		case 3:
			w->barrel.menu = -1;
			ClearTiles(w);
			break;
		case 4:
			w->barrel.menu = -1;
			UndoTiles(w);
			break;
		case 5:
			w->barrel.menu = -1;
			RandomizeTiles(w);
			break;
		case 6:
			w->barrel.menu = -1;
			SolveTiles(w);
			break;
		case 7:
			w->barrel.menu = -1;
			OrientizeTiles(w);
			break;
		case 8:
			w->barrel.menu = -1;
			PairsTiles(w);
			break;
		case 9:
			w->barrel.menu = -1;
			PracticeTiles(w);
			break;
		default:
			w->barrel.menu = -1;
			break;
		}
	}
	if (redrawTiles && !redraw && XtIsRealized(renew) && renew->core.visible) {
		EraseFrame(c);
		if (w->barrel.focus)
			DrawFrame(w, True);
		DrawAllTiles(w);
	}
	return (redraw);
}

static void
QuitBarrel(BarrelWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
DestroyBarrel(Widget old)
{
	BarrelWidget w = (BarrelWidget) old;
	int         face;

	for (face = 0; face < MAXCOLORS; face++)
		XtReleaseGC(old, w->barrel.faceGC[face]);
	XtReleaseGC(old, w->barrel.tileGC);
	XtReleaseGC(old, w->barrel.tileBrighterGC);
	XtReleaseGC(old, w->barrel.tileDarkerGC);
	XtReleaseGC(old, w->barrel.frameGC);
	XtReleaseGC(old, w->barrel.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->barrel.select);
}
#endif

#ifndef WINVER
static
#endif
void
ResizeBarrel(BarrelWidget w)
{
#ifdef WINVER
	RECT        rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom; 
#endif

	w->barrel.delta.x = 3;
	w->barrel.delta.y = 3;
	w->barrel.offset.x = MAX(((int) w->core.width - w->barrel.delta.x) /
				w->barrel.tiles, 0);
	w->barrel.offset.y = MAX(((int) w->core.height - w->barrel.delta.y) /
				w->barrel.faces, 0);
	w->barrel.faceSize.x = w->barrel.offset.x * w->barrel.tiles +
		w->barrel.delta.x;
	w->barrel.faceSize.y = w->barrel.offset.y + w->barrel.delta.y;
	w->barrel.puzzleSize.x = w->barrel.faceSize.x + w->barrel.delta.x;
	w->barrel.puzzleSize.y = (w->barrel.faceSize.y - w->barrel.delta.y) *
		w->barrel.faces + w->barrel.delta.y;
	w->barrel.puzzleOffset.x = ((int) w->core.width -
				   w->barrel.puzzleSize.x + 2) / 2;
	w->barrel.puzzleOffset.y = ((int) w->core.height -
				   w->barrel.puzzleSize.y + 2) / 2;
	w->barrel.tileSize.x = MAX(w->barrel.offset.x - w->barrel.delta.x, 0);
	w->barrel.tileSize.y = MAX(w->barrel.offset.y - w->barrel.delta.y, 0);
	ResizeTiles(w);
}

#ifndef WINVER
static
#endif
void
SizeBarrel(BarrelWidget w)
{
	ResetTiles(w);
	ResizeBarrel(w);
}

#ifndef WINVER
static
#endif
void
InitializeBarrel(
#ifdef WINVER
BarrelWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
#ifdef WINVER
        SetValuesBarrel(w);
#else
	BarrelWidget w = (BarrelWidget) renew;

	w->barrel.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->barrel.mono);
	SetAllColors(w, True);
#endif
	w->barrel.tiles = MAXTILES;
	w->barrel.faces = MAXFACES;
	w->barrel.tileOfPosition = NULL;
	CheckTiles(w);
	InitMoves();
	ResetTiles(w);
	w->barrel.numSlices = ((w->barrel.delay < MAXSLICES) ?
		w->barrel.delay + 1 : MAXSLICES);
	w->barrel.cheat = False;
	SizeBarrel(w);
#ifdef WINVER
	brush = CreateSolidBrush(w->barrel.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
#endif
}

void
ExposeBarrel(
#ifdef WINVER
BarrelWidget w
#else
Widget renew, XEvent * event, Region region
#endif
)
{
#ifndef WINVER
	BarrelWidget w = (BarrelWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
	DrawFrame(w, w->barrel.focus);
	DrawAllTiles(w);
}

#ifndef WINVER
static
#endif
void
HideBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	SetBarrel(w, BARREL_HIDE);
}

#ifndef WINVER
static
#endif
void
SelectBarrel(BarrelWidget w
#ifdef WINVER
, const int x, const int y, const int shift
#else
, XEvent * event, char **args, int nArgs
#endif
)
{
	int         i, j, pos;
#ifndef WINVER
	int x = event->xbutton.x, y = event->xbutton.y;
	int shift = (int) (event->xkey.state & (ShiftMask | LockMask));
#endif

	pos = SelectTiles(w, x, y, &i, &j);
	if (-w->barrel.tileFaces != pos) {
		w->barrel.currentTile = i;
		w->barrel.currentFace = j;
		w->barrel.currentRef = pos;
		if (shift || w->barrel.practice || !CheckSolved(w)) {
			pos = w->barrel.currentTile + w->barrel.currentFace * w->barrel.tiles;
			DrawTile(w, pos, (w->barrel.tileOfPosition[i + j * w->barrel.tiles] < 0),
				 False, TRUE, 0, 0);
		}
	} else
		w->barrel.currentRef = -w->barrel.tileFaces;
}

#ifndef WINVER
static
#endif
void
ReleaseBarrel(BarrelWidget w
#ifdef WINVER
, const int x, const int y, const int shift
#else
, XEvent * event, char **args, int nArgs
#endif
)
{
	int         i, j, pos, diff;
#ifndef WINVER
	int x = event->xbutton.x, y = event->xbutton.y;
	int shift = (int) (event->xkey.state & (ShiftMask | LockMask));
#endif

	if (w->barrel.currentRef == -w->barrel.tileFaces)
		return;
	pos = w->barrel.currentTile + w->barrel.currentFace * w->barrel.tiles;
	DrawTile(w, pos, (w->barrel.tileOfPosition[pos] < 0), True, TRUE, 0, 0);
	if (w->barrel.tileOfPosition[pos] > 0) {
		DrawTile(w, pos, False, False, FALSE, 0, 0);
	}
	if (!shift && !w->barrel.practice && CheckSolved(w)) {
		MoveNoTiles(w);
		w->barrel.currentRef = -w->barrel.tileFaces;
		return;
	}	
	pos = SelectTiles(w, x, y, &i, &j);
	if (-w->barrel.tileFaces != pos) {
		if (j == w->barrel.currentFace) {
			if ((w->barrel.spacePosition[0] % w->barrel.tiles) == 0)
				diff = w->barrel.tiles - 1;
			else
				diff = -w->barrel.tiles + 1;
			(void) SelectSlideTiles(w, diff);
			if (CheckSolved(w)) {
				SetBarrel(w, BARREL_SOLVED);
			}
		} else {
			if (!shift && (w->barrel.currentTile == 0 ||
					w->barrel.currentTile == w->barrel.tiles - 1)) {
				w->barrel.currentRef = -w->barrel.tileFaces;
				return;
			}
			diff = (w->barrel.currentFace - j + w->barrel.faces) % w->barrel.faces;
			if (diff > w->barrel.faces / 2)
				for (i = 0; i < w->barrel.faces - diff; i++)
					(void) MoveBarrel(w, BOTTOM, w->barrel.currentTile, shift, NORMAL);
			else
				for (i = 0; i < diff; i++)
					(void) MoveBarrel(w, TOP, w->barrel.currentTile, shift, NORMAL);
			if (!shift && CheckSolved(w)) {
				SetBarrel(w, BARREL_SOLVED);
			}
		}
	}
	w->barrel.currentRef = -w->barrel.tileFaces;
}

#ifndef WINVER
static
#endif
void
PracticeBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	PracticeTiles(w);
}

#ifndef WINVER
static
#endif
void
PracticeBarrelMaybe(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	if (!w->barrel.started)
		PracticeTiles(w);
#ifdef HAVE_MOTIF
	else {
		SetBarrel(w, BARREL_PRACTICE_QUERY);
	}
#endif
}

#ifndef WINVER
static
#endif
void
PracticeBarrel2(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
#ifdef HAVE_MOTIF
	if (!w->barrel.started)
#endif
		PracticeTiles(w);
}

#ifndef WINVER
static
#endif
void
RandomizeBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	RandomizeTiles(w);
}

#ifndef WINVER
static
#endif
void
RandomizeBarrelMaybe(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	if (!w->barrel.started)
		RandomizeTiles(w);
#ifdef HAVE_MOTIF
	else {
		SetBarrel(w, BARREL_RANDOMIZE_QUERY);
	}
#endif
}

#ifndef WINVER
static
#endif
void
RandomizeBarrel2(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
#ifdef HAVE_MOTIF
	if (!w->barrel.started)
#endif
		RandomizeTiles(w);
}

#ifndef WINVER
static
#endif
void
GetBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	GetTiles(w);
}

#ifndef WINVER
static
#endif
void
WriteBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	WriteTiles(w);
}

#ifndef WINVER
static
#endif
void
ClearBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	ClearTiles(w);
}

#ifndef WINVER
static
#endif
void
UndoBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	UndoTiles(w);
}

#ifndef WINVER
static
#endif
void
SolveBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	SolveTiles(w);
}

#ifndef WINVER
static
#endif
void
OrientizeBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	OrientizeTiles(w);
}

#ifndef WINVER
static
#endif
void
PairsBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	PairsTiles(w);
}

#ifndef WINVER
static
#endif
void
EnterBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	w->barrel.focus = True;
	DrawFrame(w, w->barrel.focus);
}

#ifndef WINVER
static
#endif
void
LeaveBarrel(BarrelWidget w
#ifndef WINVER
, XEvent * event, char **args, int nArgs
#endif
)
{
	w->barrel.focus = False;
	DrawFrame(w, w->barrel.focus);
}

#ifndef WINVER
static void
MoveBarrelTop(BarrelWidget w, XEvent * event, char **args, int nArgs)
{
	MoveBarrelInput(w, event->xbutton.x, TOP,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveBarrelLeft(BarrelWidget w, XEvent * event, char **args, int nArgs)
{
	MoveBarrelInput(w, event->xbutton.x, LEFT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveBarrelRight(BarrelWidget w, XEvent * event, char **args, int nArgs)
{
	MoveBarrelInput(w, event->xbutton.x, RIGHT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveBarrelBottom(BarrelWidget w, XEvent * event, char **args, int nArgs)
{
	MoveBarrelInput(w, event->xbutton.x, BOTTOM,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}
#endif
