/*-
# X-BASED CUBES
#
#  CubesS.c
#
###
#
#  Taken from the X puzzle by Don Bennett, HP Labs
#
#  Copyright (c) 2003    David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  TODO
#   o handle < 4x4 and 3d
#
*/

/*
 *	$XConsortium: puzzle.c,v 1.14 94/03/28 18:34:10 gildea Exp $
 */

/* Puzzle - (C) Copyright 1987, 1988 Don Bennett.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 */

/**
 **  Puzzle
 **
 ** Don Bennett, HP Labs
 **
 ** this is the code that does the real work to solve the
 ** puzzle.  (Commonly seen as a 4x4 grid of sliding pieces
 ** numbered 1-15 with one empty space.)
 **
 ** The idea for the solution algorithm - solving the puzzle
 ** in layers working from the outside in - comes to me
 ** indirectly from John Nagle.
 **/

#ifndef WINVER
#define JMP
#ifdef JMP
#include <setjmp.h> /* longjmp ... interrupt */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#endif
#include "CubesP.h"
#define MAX_PLAN	1000

/** PuzzleSize MUST be a multiple of 2; **/
#define PuzzleWidth w->cubes.sizeX
#define PuzzleHeight w->cubes.sizeY
#define PuzzleRect w->cubes.sizeRect
#define PuzzleSize (MIN((PuzzleWidth/2)*2,(PuzzleHeight/2)*2))
#define position w->cubes.blockOfPosition
#define space_x	(w->cubes.spacePosition%w->cubes.sizeX)
#define space_y	(w->cubes.spacePosition/w->cubes.sizeX)

static int other_dir[4] = { BOTTOM, LEFT, TOP, RIGHT };

/** layer info macros ->  (innermost 4 tiles are layer zero, ordinal goes up
 **			   as you move out)
 ** layer_depth		- returns number of (rows down),(cols across) the layer starts;
 ** layer_width   	- number of blocks wide the layer is;
 **/

#define layer_depth(l)	(layers-1-(l))
#define layer_width(l)	(PuzzleSize-2*layer_depth(l))

/** macros for finding the corners of each layer **/

#define UL(l)	(layer_depth(l)*(PuzzleSize+1)+\
  ExtraRows*PuzzleWidth + ExtraColumns*(layers-(l)))
#define UR(l)   (layer_depth(l)*(PuzzleSize+1)+layer_width(l)-1+\
  ExtraRows*PuzzleWidth+ExtraColumns*(layers-(l)))
#define LL(l)   ((layer_depth(l)+layer_width(l)-1)*PuzzleSize+layer_depth(l)+\
  ExtraRows*PuzzleSize+ExtraColumns*(PuzzleHeight+1+(l)-layers))
#define LR(l)	((layer_depth(l)+layer_width(l)-1)*(PuzzleSize+1)+\
  ExtraRows*PuzzleSize+ExtraColumns*(PuzzleHeight+1+(l)-layers))

/** get the x and y coordinates of a location in the matrix **/

#define get_x(loc)	((loc)%PuzzleWidth)
#define get_y(loc)	((loc)/PuzzleWidth)
#define indx(x,y)	(((y)*PuzzleWidth)+(x))

#define next_left(loc)	((loc)-1)
#define next_right(loc)	((loc)+1)
#define next_up(loc)	((loc)-PuzzleWidth)
#define next_down(loc)	((loc)+PuzzleWidth)

static int sizex = 0, sizey = 0;
static int ExtraRows = 0;
static int ExtraColumns = 0;
static int layers;
static int *tmp_matrix;
static int *targetm;
static int *locked;
static int *loclist;

static Boolean SolvingFlag = False;
static Boolean AbortSolvingFlag = False;
/* TODO Interruptability for Windows */
#ifdef JMP
static jmp_buf solve_env;

static void
AbortSolving(void)
{
	if (SolvingFlag)
		AbortSolvingFlag = True;
}

static void
ProcessButton(void /*XButtonEvent *event*/)
{
	AbortSolving();
}

static void
ProcessVisibility(XVisibilityEvent *event)
{
	if (event->state != VisibilityUnobscured)
		AbortSolving();
}

static void
GetNextEvent(CubesWidget w, XEvent *event)
{
	if (!XCheckMaskEvent(XtDisplay(w), VisibilityChangeMask, event))
		(void) XNextEvent(XtDisplay(w), event);
}

static void
ProcessEvent(XEvent *event)
{
	switch(event->type) {
		case KeyPress:
		case ButtonPress:
			ProcessButton(/*&event->xbutton*/);
			break;
		case VisibilityNotify:
			ProcessVisibility(&event->xvisibility);
			break;
		default:
			break;
	}
}

static void
ProcessEvents(CubesWidget w)
{
	XEvent event;

	while(XPending(XtDisplay(w))) {
		GetNextEvent(w, &event);
		ProcessEvent(&event);
	}
}
#endif

static int
find_piece(CubesWidget w, int piece)
{
	int i;
	char *buf1 = NULL, *buf2 = NULL;

	for (i = 0; i < PuzzleRect; i++)
		if (position[i] == piece)
			return (i);

	intCat(&buf1, "Piece ", piece);
	stringCat(&buf2, buf1, "not found, exiting.");
	free(buf1);
	DISPLAY_ERROR(buf2);
	free(buf2);
	return -1;
}

static void
LogMoveSpace(CubesWidget w,
		int first_x, int first_y, int last_x, int last_y, int dir)
{
	int count = ABS(first_x - last_x) + ABS(first_y - last_y);

	AnimateSlide(w, count + 1, dir, NORMAL, True);
}

static void
move_space(CubesWidget w, int dir, int dist)
{
	int i, step, count;
	int first_x, first_y;
	int last_x, last_y, shift_dir;
	int temp_space_x = space_x, temp_space_y = space_y;

#ifdef JMP
	if (XPending(XtDisplay(w)))
		ProcessEvents(w);
	if (SolvingFlag && AbortSolvingFlag)
		longjmp(solve_env, 1);
#endif
	if (dist == 0)
		return;

	if (dir == LEFT) {
		dir = RIGHT;
		dist = -dist;
	}
	if (dir == TOP) {
		dir = BOTTOM;
		dist = -dist;
	}
	first_x = space_x;
	first_y = space_y;
	step = 1;
	count = dist;
	if (dist < 0) {
		step  = -1;
		count = -count;
	}

	/** first_x,y are the location of the first piece to be shifted **/
	if (dir == RIGHT)
		first_x += step;
	else
		first_y += step;

	/** shift_dir is the direction the pieces need to be shifted **/
	if (dist < 0) {
		shift_dir = dir;
	} else {
		shift_dir= other_dir[dir];
	}
	for (i = 0; i < count; i++) {
		if (dir == RIGHT) {
			temp_space_x += step;
		} else { /** dir == BOTTOM **/
			temp_space_y += step;
		}
	}
	last_x = temp_space_x;
	last_y = temp_space_y;

	/** the blocks first_x,y through last_x,y need to be shifted
	 ** one block in the shift_dir direction;
	 **/

	LogMoveSpace(w, first_x, first_y, last_x, last_y, shift_dir);
}

#ifdef DEBUG
static void
print_matrix(CubesWidget w, int **mat)
{
	int i, j;

	(void) printf("\n");
	for (j = 0; j < PuzzleHeight; j++) {
		for (i = 0; i < PuzzleWidth; i++)
			(void) printf(" %2d ",(*mat)[indx(i, j)]);
		(void) printf("\n");
	}
	(void) printf("\n");
}
#endif

static void
plan_move(CubesWidget w, int start_loc, int end_loc, int **path)
{

#define QUEUE_SIZE	1000
#define DIST(loc1,loc2)	(ABS(get_x(loc1)-get_x(loc2))+\
  ABS(get_y(loc1)-get_y(loc2)))

   int i, next_loc = 0, next_dist, chosen, found_path, move_num;
   int loc_x, loc_y;
   int loc_queue[QUEUE_SIZE];
   int loc_dist[QUEUE_SIZE];
   int loc_queue_used[QUEUE_SIZE];
   int queue_head, queue_tail;
   int candidate[4];

   found_path = 0;

   for (i = 0; i < PuzzleRect; i++) {
      tmp_matrix[i] = -locked[i];
      loclist[i] = -1;
   }

   for (i = 0; i < QUEUE_SIZE; i++)
      loc_queue_used[i] = 0;

   queue_head = 0;
   queue_tail = 0;

   loc_queue[0] = start_loc;
   loc_dist[0]  = DIST(end_loc, start_loc);
   tmp_matrix[start_loc] = 1;
   queue_tail++;

   /** if the selected element has a distance of zero, we've found it;
    **  (This really isn't a queue, but rather a range of elements
    ** to be searched for an element of the desired properties;
    **/

   /** as we search for a path,
    ** LINK       array is used to indicate the direction from which
    **            we moved into a location;
    ** TMP_MATRIX array is used to keep track of the move number;
    **/

   while (queue_head < queue_tail && !found_path) {
      /** find the entry that
       ** (1) has the smallest distance and
       ** (2) has the smallest move number;
       **/

      next_loc = loc_queue[queue_head];
      next_dist = loc_dist[queue_head];
      chosen = queue_head;

      for (i = queue_head + 1; i < queue_tail; i++)
         if (!loc_queue_used[i] && ((loc_dist[i] < next_dist) ||
                 ((loc_dist[i] == next_dist) &&
                 (tmp_matrix[loc_queue[i]] < tmp_matrix[next_loc])))) {
            next_loc = loc_queue[i];
            next_dist = loc_dist[i];
            chosen = i;
         }

      if (next_dist == 0) {
         found_path = 1;
         break;
      }

      loc_queue_used[chosen] = 1;

      /********************************/
      /** permute the chosen element **/
      /********************************/

      candidate[TOP] = next_up(next_loc);
      candidate[LEFT] = next_left(next_loc);
      candidate[BOTTOM] = next_down(next_loc);
      candidate[RIGHT] = next_right(next_loc);

      loc_x = get_x(next_loc);
      loc_y = get_y(next_loc);

      if (loc_x == 0)
	      candidate[LEFT] = -1;
      if (loc_x == PuzzleWidth - 1)
	      candidate[RIGHT] = -1;
      if (loc_y == 0)
	      candidate[TOP] = -1;
      if (loc_y == PuzzleHeight - 1)
	      candidate[BOTTOM] = -1;

      move_num = tmp_matrix[next_loc] + 1;

      for (i = 0; i < 4; i++)
         if (candidate[i] != -1 && tmp_matrix[candidate[i]] == 0) {
            tmp_matrix[candidate[i]] = move_num;
            /** the next line works because the candidate index is
             ** same as the direction moved to reach the candidate;
             **/
            loclist[candidate[i]] = i;
            loc_queue[queue_tail] = candidate[i];
            loc_dist[queue_tail] = DIST(end_loc, candidate[i]);
            queue_tail++;
            if (queue_tail == QUEUE_SIZE)
		goto broke;
         }

      /***************************************************/
      /** delete used items from the front of the queue **/
      /***************************************************/

      while (loc_queue_used[queue_head] && queue_head < queue_tail)
         queue_head++;
   }

#ifdef DEBUG
      print_matrix(w, &position);
      (void) printf("\n");
      print_matrix(w, &locked);
      (void) printf("\n");
#endif /* DEBUG */
   if (!found_path) {
	char *buf1 = NULL, *buf2 = NULL;

	intCat(&buf1, "Could not find a way to move ", start_loc);
	stringCat(&buf2, buf1, " (");
	free(buf1);
	intCat(&buf1, buf2, get_x(start_loc));
	free(buf2);
	stringCat(&buf2, buf1, ",");
	free(buf1);
	intCat(&buf1, buf2, get_y(start_loc));
	free(buf2);
	stringCat(&buf2, buf1, ") to ");
	free(buf1);
	intCat(&buf1, buf2, end_loc);
	free(buf2);
	stringCat(&buf2, buf1, " (");
	free(buf1);
	intCat(&buf1, buf2, get_x(end_loc));
	free(buf2);
	stringCat(&buf2, buf1, ",");
	free(buf1);
	intCat(&buf1, buf2, get_y(end_loc));
	free(buf2);
	stringCat(&buf2, buf1, "), exiting.");
	free(buf1);
	DISPLAY_ERROR(buf2);
	free(buf2);
	return;
   }

broke:	if (queue_tail == QUEUE_SIZE) {
		DISPLAY_ERROR("Queue size is exceeded, exiting.");
		return;
	}

   /** copy the path we found into the path array;
    ** element 0 will contain the number of moves in the path;
    **/

   /** by the time we get there, next_loc is in the final location **/

   (*path)[0] = tmp_matrix[next_loc] - 1;
   for (i = (*path)[0]; i > 0; i--) {
      (*path)[i] = loclist[next_loc];
      switch(loclist[next_loc]) {
      case LEFT:	next_loc = next_right(next_loc);
			break;
      case RIGHT:	next_loc = next_left(next_loc);
			break;
      case TOP:		next_loc = next_down(next_loc);
			break;
      case BOTTOM:	next_loc = next_up(next_loc);
			break;
      }
   }
}

static void
my_move_space_to(CubesWidget w, int loc)
{
	int i, current_dir, dist;
	int plan[MAX_PLAN];
	int *planPtr = &(plan[0]);

	plan_move(w, indx(space_x, space_y), loc, &planPtr);
	current_dir = plan[1];
	dist = 0;
	for (i = 1; i <= plan[0]; i++) {
		if (plan[i] == current_dir)
			dist++;
		else if (plan[i] == other_dir[current_dir])
			dist--;
		else {
			move_space(w, current_dir, dist);
			current_dir = plan[i];
			dist = 1;
		}
	}
	move_space(w, current_dir, dist);
}

static void
move_piece(CubesWidget w, int loc, int target_m)
{
   int i;
   int plan[MAX_PLAN];
   int *planPtr = &(plan[0]);

   plan_move(w, loc, target_m, &planPtr);
   for (i = 1; i <= plan[0]; i++)
      switch(plan[i]) {
      case LEFT:	locked[loc] = 1;
			my_move_space_to(w, next_left(loc));
			locked[loc] = 0;
			my_move_space_to(w, loc);
			loc = next_left(loc);
			break;
      case RIGHT:	locked[loc] = 1;
			my_move_space_to(w, next_right(loc));
			locked[loc] = 0;
			my_move_space_to(w, loc);
			loc = next_right(loc);
			break;
      case TOP:		locked[loc] = 1;
			my_move_space_to(w, next_up(loc));
			locked[loc] = 0;
			my_move_space_to(w, loc);
			loc = next_up(loc);
			break;
      case BOTTOM:	locked[loc] = 1;
			my_move_space_to(w, next_down(loc));
			locked[loc] = 0;
			my_move_space_to(w, loc);
			loc = next_down(loc);
			break;
      }
}

/* SYSV386 gets this from libBerk.a */
#if defined(USG) && !defined(CRAY) && !defined(SYSV386)
int gettimeofday (tvp, tzp)
    struct timeval *tvp;
    struct timezone *tzp;
{
    time (&tvp->tv_sec);
    tvp->tv_usec = 0L;

    /* ignore tzp for now since this file doesn't use it */
}
#endif

static void
myFree(void)
{
	if (tmp_matrix)
		free(tmp_matrix);
	if (targetm)
		free(targetm);
	if (locked)
		free(locked);
	if (loclist)
		free(loclist);
}

static void
myInitialize(CubesWidget w)
{
   /** Initialize the position and
    ** the targetm matrices;
    **/

   int i;
   int sp_x, sp_y;

sizex=PuzzleWidth;
sizey=PuzzleHeight;
   layers = PuzzleSize >> 1;

   ExtraRows    = PuzzleHeight - PuzzleSize;
   ExtraColumns = PuzzleWidth - PuzzleSize;

   tmp_matrix = (int *) malloc(PuzzleRect * sizeof(int));
   targetm     = (int *) malloc(PuzzleRect * sizeof(int));
   locked     = (int *) malloc(PuzzleRect * sizeof(int));
   loclist    = (int *) malloc(PuzzleRect * sizeof(int));

   for (i = 0; i < PuzzleRect; i++)
      locked[i] = 0;

   if (!tmp_matrix || !targetm || !locked || !loclist || !position) {
       DISPLAY_ERROR("Not enough memory, exiting.");
   }

   for (i = 0; i < PuzzleRect - 1; i++) {
      targetm[i] = i + 1;
   }
   targetm[PuzzleRect - 1] = 0;

   /** Move the space into the LR corner of the
    ** innermost layer;
    ** For each of the outer layers, move the space
    ** left one and up one;
    **/

   sp_x = PuzzleWidth - 1;
   sp_y = PuzzleHeight - 1;

   /* Make solution with space at center bottom right */
   for (i = 0; i < layers - 1; i++) {
      /** move the space left one; **/
      targetm[indx(sp_x, sp_y)] = targetm[indx((sp_x - 1), sp_y)];
      targetm[indx((sp_x - 1),sp_y)] = 0;
      sp_x -= 1;

      /** move the space up one; **/
      targetm[indx(sp_x, sp_y)] = targetm[indx(sp_x, (sp_y - 1))];
      targetm[indx(sp_x, (sp_y - 1))] = 0;
      sp_y -= 1;
   }
}

/** To solve this puzzle, work from the outside in;
 ** For each successive ring working your way in,
 **
 ** (1) put the corners in place;
 ** (2) finish off the rest of the boundaries;
 ** (3) do the next layer in;
 **/

static void
solve_layer_0(CubesWidget w)
{
	move_piece(w, find_piece(w, targetm[UL(0)]), UL(0));
	my_move_space_to(w, LR(0));
}

static void
do_last_two_on_edge(CubesWidget w, int ntlast, int last, int tmp, int emergency)
{
   int last_piece, ntlast_piece;
   last_piece = targetm[last];
   ntlast_piece = targetm[ntlast];

   move_piece(w, find_piece(w, ntlast_piece),last);
   locked[last] = 1;

   /** if the last piece is stuck where the next to the last
    ** piece should go, do some magic to fix things up;
    **/
   if (find_piece(w, 0) == ntlast)
      my_move_space_to(w, tmp);

   if (find_piece(w, last_piece) == ntlast) {
   /** a rescue is necessary **/
      locked[last] = 0;
      move_piece(w, find_piece(w, ntlast_piece), ntlast);
      locked[ntlast] = 1;
      move_piece(w, find_piece(w, last_piece), emergency);
      locked[emergency] = 1;
      locked[ntlast] = 0;
      move_piece(w, find_piece(w, ntlast_piece), last);
      locked[emergency] = 0;
      locked[last] = 1;
   }

   move_piece(w, find_piece(w, last_piece), tmp);
   locked[tmp] = 1;
   my_move_space_to(w, ntlast);
   locked[tmp]  = 0;
   locked[last] = 0;
   my_move_space_to(w, last);
   my_move_space_to(w, tmp);
   locked[ntlast] = 1;
   locked[last] = 1;
}

static void
solve_layer(CubesWidget w, int layer)
{
   int i, tmp, last, ntlast, emergency;
   int ul, ur, ll, lr;

   if (layer == 0) {
      solve_layer_0(w);
   } else {
      /** find and put each of the corners into place **/
      ul = UL(layer);
      ur = UR(layer);
      ll = LL(layer);
      lr = LR(layer);

      move_piece(w, find_piece(w, targetm[ul]), ul);
      locked[ul] = 1;
      move_piece(w, find_piece(w, targetm[ur]), ur);
      locked[ur] = 1;
      move_piece(w, find_piece(w, targetm[ll]), ll);
      locked[ll] = 1;
      move_piece(w, find_piece(w, targetm[lr]), lr);
      locked[lr] = 1;

      /** Strategy for doing the pieces between the corners:
       ** (1) put all but the last two edge pieces in place;
       ** (2) put the next to the last piece next to the corner;
       ** (3) put the last piece one move in from its final position;
       ** (4) move the space to the final position of the next
       **     to the last piece;
       ** (5) slide the next to the last piece over and the last
       **     piece into the edge where it goes.
       **/

      /**************/
      /** top edge **/
      /**************/
      for (i = ul + 1; i < ur - 2; i++) {
         move_piece(w, find_piece(w, targetm[i]), i);
         locked[i] = 1;
      }

      ntlast = i;
      last   = i + 1;
      tmp    = UR(layer - 1);
      emergency = next_down(tmp);
      do_last_two_on_edge(w, ntlast, last, tmp, emergency);

      /*****************/
      /** bottom edge **/
      /*****************/
      for (i = ll + 1; i < lr - 2; i++) {
         move_piece(w, find_piece(w, targetm[i]), i);
         locked[i] = 1;
      }

      ntlast = i;
      last   = i + 1;
      tmp    = LR(layer - 1);
      emergency = next_up(tmp);
      do_last_two_on_edge(w, ntlast, last, tmp, emergency);

      /***************/
      /** left side **/
      /***************/
      for (i = ul + PuzzleWidth; i < ll - 2 * PuzzleWidth; i += PuzzleWidth) {
         move_piece(w, find_piece(w, targetm[i]), i);
         locked[i] = 1;
      }

      ntlast = i;
      last   = i + PuzzleWidth;
      tmp    = LL(layer - 1);
      emergency = next_right(tmp);
      do_last_two_on_edge(w, ntlast, last, tmp, emergency);

      /****************/
      /** right side **/
      /****************/
      for (i = ur + PuzzleWidth; i < lr - 2 * PuzzleWidth; i += PuzzleWidth) {
         move_piece(w, find_piece(w, targetm[i]), i);
         locked[i] = 1;
      }

      ntlast = i;
      last   = i + PuzzleWidth;
      tmp    = LR(layer - 1);
      emergency = next_left(tmp);
      do_last_two_on_edge(w, ntlast, last, tmp, emergency);
   }
}

static void
solve_row(CubesWidget w, int row)
{
	int i, loc, last, ntlast, tmp, emergency;

	for (i = 0; i < PuzzleWidth - 2; i++) {
		loc = indx(i, row);
		move_piece(w, find_piece(w, targetm[loc]), loc);
		locked[loc] = 1;
	}
	ntlast = indx(PuzzleWidth - 2, row);
	last = indx(PuzzleWidth - 1, row);
	tmp = last + PuzzleWidth;
	emergency = tmp + PuzzleWidth;
	do_last_two_on_edge(w, ntlast, last, tmp, emergency);
}

static void
solve_col(CubesWidget w, int col)
{
	int i, loc, last, ntlast, tmp, emergency;

	for (i = 0; i < PuzzleHeight - 2; i++) {
		loc = indx(col, i);
		move_piece(w, find_piece(w, targetm[loc]), loc);
		locked[loc] = 1;
	}
	ntlast = indx(col, PuzzleHeight - 2);
	last = indx(col, PuzzleHeight - 1);
	tmp = last + 1;
	emergency = tmp + 1;
	do_last_two_on_edge(w, ntlast, last, tmp, emergency);
}

/* This procedure coordinates the solution process. */
void
SolveSomeBlocks(CubesWidget w)
{
	int i;

	if (sizex != w->cubes.sizeX || sizey != w->cubes.sizeY) {
		myFree();
		myInitialize(w);
	}
        SetCubes(w, CUBES_RESET);
	/** determine the position we want to be in when
	 ** we are done; This position will have the space in
	 ** the center;  Then, we'll move the space back to
	 ** the outside.
	 **/
	if (SolvingFlag)
		return;

#ifdef JMP
	if (!setjmp(solve_env))
#else
	if (1)
#endif
	{
		SolvingFlag = True;
		for (i = 0; i < PuzzleRect; i++)
			locked[i] = 0;
		/** solve the extra rows and cols **/
		for (i = 0; i < ExtraRows; i++)
			solve_row(w, i);
		for (i = 0; i < ExtraColumns; i++)
			solve_col(w, i);

		/** solve each layer **/
		for (i = layers - 1; i >= 0; i--) {
			solve_layer(w, i);
		}
		/** move the space back out to the LR corner; **/
		/** i is the layer the space is moving into **/
		for (i = 1; i < layers; i++) {
			move_space(w, BOTTOM, 1);
			move_space(w, RIGHT, 1);
		}
	} else {
		DrawAllBlocks(w);
	}
	for (i = 0; i < PuzzleRect; i++)
		locked[i] = 0;
	AbortSolvingFlag = False;
	SolvingFlag = False;
	w->cubes.cheat = True; /* Assume the worst. */
        SetCubes(w, CUBES_COMPUTED);
}
