/*-
# X-BASED HEXAGONS
#
#  HexagonP.h
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Hexagons */

#ifndef _HexagonsP_h
#define _HexagonsP_h

#include "Hexagons.h"

#ifdef WINVER
#define STRICT

#include <windows.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if ((WINVER > 0x030a) && !defined(GCL_HBRBACKGROUND))
#undef WINVER
#define WINVER 0x030a
#endif
#if (WINVER <= 0x030a)          /* if WINDOWS 3.1 or less */
#define POS int
#define SETBACK(h,b) (void)SetClassWord(h,GCW_HBRBACKGROUND,(WORD)b);
#define MoveTO(h,x,y,lp) MoveTo(h,x,y);
#else
#define POS long
#define SETBACK(h,b) (void)SetClassLong(h,GCL_HBRBACKGROUND,(LONG)b);
#define MoveTO(h,x,y,lp) MoveToEx(h,x,y,lp);
#endif
#ifndef Point
#define Point POINT
#endif
#ifndef Boolean
#define Boolean BOOL
#endif
#ifndef True
#define True TRUE
#endif
#ifndef False
#define False FALSE
#endif
#define GC COLORREF  /* Equate apples to oranges */

#define RECTANGLE(w,x,y,l,h) (void)Rectangle(w->core.hDC,x,y,x+l,y+h)
#define LINE(w,x1,y1,x2,y2) (void)MoveTO(w->core.hDC,x1,y1,NULL); \
  (void)LineTo(w->core.hDC,x2,y2)
#define CIRCLE(w,r,x,y) if(r>0)(void)Ellipse(w->core.hDC,x-r+1,y-r+1,x+r,y+r)
#define DRAWRECTANGLE(w,c,x,y,l,h) hPen = CreatePen(PS_SOLID, 1, c); \
  hOldPen = (HPEN) SelectObject(w->core.hDC, hPen); \
  RECTANGLE(w,x,y,l,h); \
  (void) SelectObject(w->core.hDC, hOldPen); \
  (void) DeleteObject(hPen)
#define FILLRECTANGLE(w,c,x,y,l,h) hPen = CreatePen(PS_SOLID, 1, c); \
  hOldPen = (HPEN) SelectObject(w->core.hDC, hPen); \
  hBrush = CreateSolidBrush(c); \
  hOldBrush = (HBRUSH) SelectObject(w->core.hDC, hBrush); \
  RECTANGLE(w,x,y,l,h); \
  (void) SelectObject(w->core.hDC, hOldBrush); \
  (void) DeleteObject(hBrush); \
  (void) SelectObject(w->core.hDC, hOldPen); \
  (void) DeleteObject(hPen)
#define DRAWLINE(w,c,x1,y1,x2,y2) hPen = CreatePen(PS_SOLID, 1, c); \
  hOldPen = (HPEN) SelectObject(w->core.hDC, hPen); \
  LINE(w,x1,y1,x2,y2); \
  (void) SelectObject(w->core.hDC, hOldPen); \
  (void) DeleteObject(hPen)
#define DRAWCIRCLE(w,c,d,x,y) hPen = CreatePen(PS_SOLID, 1, c); \
  hOldPen = (HPEN) SelectObject(w->core.hDC, hPen); \
  CIRCLE(w,((d+1)>>1),x,y); \
  (void) SelectObject(w->core.hDC, hOldPen); \
  (void) DeleteObject(hPen)
#define FILLCIRCLE(w,c,d,x,y) hPen = CreatePen(PS_SOLID, 1, c); \
  hOldPen = (HPEN) SelectObject(w->core.hDC, hPen); \
  hBrush = CreateSolidBrush(c); \
  hOldBrush = (HBRUSH) SelectObject(w->core.hDC, hBrush); \
  CIRCLE(w,((d+1)>>1),x,y); \
  (void) SelectObject(w->core.hDC, hOldBrush); \
  (void) DeleteObject(hBrush); \
  (void) SelectObject(w->core.hDC, hOldPen); \
  (void) DeleteObject(hPen)
#define DRAWTEXT(w,c,x,y,s,l) (void) SetTextColor(w->core.hDC, c); \
  (void) SetBkMode(w->core.hDC, TRANSPARENT); \
  (void) TextOut(w->core.hDC, (x)-3, (y)-11, (LPCSTR) s, l)

#if 0
#define DISPLAY_WARNING(s) (void) fprintf(stderr, s)
#else
#define DISPLAY_WARNING(s) (void) MessageBox(w->core.hWnd, (LPCSTR) s, \
  "Warning", MB_OK)
#endif
#define DISPLAY_ERROR(s) (void) fprintf(stderr, "%s\n", s); exit(1)

typedef struct _CorePart {
	HWND        hWnd;
	HDC         hDC;
	HINSTANCE   hInstance;
	POS         width, height;
} CorePart;

extern HBRUSH hOldBrush, hBrush;
extern HPEN hOldPen, hPen;

extern void DestroyHexagons(HBRUSH brush);
extern void ResizeHexagons(HexagonsWidget w);
extern void SizeHexagons(HexagonsWidget w);
extern void InitializeHexagons(HexagonsWidget w, HBRUSH brush);
extern void ExposeHexagons(const HexagonsWidget w);
extern void HideHexagons(const HexagonsWidget w);
extern void SelectHexagons(HexagonsWidget w, const int x, const int y);
extern void ReleaseHexagons(HexagonsWidget w, const int x, const int y);
extern void RandomizeHexagons(HexagonsWidget w);
extern void RandomizeHexagonsMaybe(HexagonsWidget w);
extern void RandomizeHexagons2(HexagonsWidget w);
extern void ClearHexagons(HexagonsWidget w);
extern void GetHexagons(HexagonsWidget w);
extern void WriteHexagons(HexagonsWidget w);
extern void UndoHexagons(HexagonsWidget w);
extern void SolveHexagons(HexagonsWidget w);
extern void ModeHexagons(HexagonsWidget w);
extern void EnterHexagons(HexagonsWidget w);
extern void LeaveHexagons(HexagonsWidget w);
extern int  MoveHexagons(HexagonsWidget w, const int direction, const int control);

#else

#ifndef Point
#define Point XPoint
#endif
#define NUM_DEGREES 360
#define MULT 64
#define FULLCIRCLE (NUM_DEGREES*MULT)
#define FILLRECTANGLE(w,c,i,j,l,h) XFillRectangle(XtDisplay(w),XtWindow(w),\
  c,i,j,l,h)
#define DRAWRECTANGLE(w,c,i,j,l,h) XDrawRectangle(XtDisplay(w),XtWindow(w),\
  c,i,j,l,h)
#define DRAWLINE(w,c,x1,y1,x2,y2) XDrawLine(XtDisplay(w),XtWindow(w),\
  c,x1,y1,x2,y2)
#define FILLCIRCLE(w,c,d,x,y) if(d>0) XFillArc(XtDisplay(w),XtWindow(w),\
  c,x-((d)>>1),y-((d)>>1),d,d,0,FULLCIRCLE)
#define DRAWCIRCLE(w,c,d,x,y) if(d>0) XDrawArc(XtDisplay(w),XtWindow(w),\
  c,x-((d)>>1),y-((d)>>1),d,d,0,FULLCIRCLE)
#define DRAWTEXT(w,c,x,y,s,l) (void) XDrawString(XtDisplay(w),XtWindow(w),\
  c,x,y,s,l)
#define POLYGON(w,c,lc,l,n,o) XFillPolygon(XtDisplay(w), XtWindow(w), c, \
  l, n, Convex, (o) ? CoordModeOrigin : CoordModePrevious); \
  XDrawLines(XtDisplay(w), XtWindow(w), lc, \
  l, (n)+1,(o) ? CoordModeOrigin : CoordModePrevious)
#define POLYLINE(w,c,l,n,o) XDrawLines(XtDisplay(w), XtWindow(w), c, \
  l, n, (o) ? CoordModeOrigin : CoordModePrevious)

#define DISPLAY_WARNING(s) XtWarning(s)
#define DISPLAY_ERROR(s) XtError(s)

/* This gets around C's inability to do inheritance */
typedef struct _HexagonsClassPart {
	int         ignore;
} HexagonsClassPart;

typedef struct _HexagonsClassRec {
	CoreClassPart core_class;
	HexagonsClassPart hexagons_class;
} HexagonsClassRec;

extern HexagonsClassRec hexagonsClassRec;
#endif

#define SYMBOL ':'

#define TR 0
#define RIGHT 1
#define BR 2
#define BL 3
#define LEFT 4
#define TL 5
#define COORD 6
#define TOP 6
#define BOTTOM 9
#define TRBL 0
#define TLBR 1
#define ROW 2
#define ROWTYPES 3
#define LOW 0
#define HIGH 1
#define SPACES 2

#define ABS(a) (((a)<0)?(-(a)):(a))
#define SIGN(a) (((a)<0)?(-1):1)
#define MIN(a,b) (((int)(a)<(int)(b))?(int)(a):(int)(b))
#define MAX(a,b) (((int)(a)>(int)(b))?(int)(a):(int)(b))

/*** random number generator ***/
#if HAVE_RAND48
#define SRAND srand48
#define LRAND lrand48
#else /* HAVE_RAND48 */
#if HAVE_RANDOM
#define SRAND srandom
#define LRAND lrandom
#else /* HAVE_RANDOM */
#if HAVE_RAND
#define SRAND srand
#define LRAND lrand
#endif /* HAVE_RAND */
#endif /* HAVE_RANDOM */
#endif /* HAVE_RAND48 */
#ifndef SRAND
extern void SetRNG(long s);
#define SRAND(X) SetRNG((long) X)
#endif
#ifndef LRAND
extern long LongRNG(void);
#define LRAND() LongRNG()
#endif
#define NRAND(X) ((int)(LRAND()%(X)))

typedef struct _HexagonsPart {
	int        *tileOfPosition, spacePosition[SPACES];
	int         spaceRow[ROWTYPES];
	int         currentPosition, currentRow[ROWTYPES];
	Boolean     started, cheat;
	Boolean     corners, mono, reverse, focus;
	int         base;
	int         sizeX, sizeY, sizeSize, sizeCenter;
	Point       offset, tileSize, puzzleSize;
	Point       delta, puzzleOffset, digitOffset;
	GC          frameGC, borderGC, inverseGC;
	GC          tileGC, tileBrighterGC, tileDarkerGC;
#ifdef WINVER
	char        username[81];
#else
	int         menu;
	char       *username;
	Pixel       foreground, background;
	Pixel       frameColor, tileColor, borderColor;
	XtCallbackList select;
#endif
} HexagonsPart;

typedef struct _HexagonsRec {
	CorePart    core;
	HexagonsPart hexagons;
} HexagonsRec;

extern int *startPosition;
extern void SetHexagons(HexagonsWidget w, int reason);
extern int  MoveHexagonsDir(HexagonsWidget w, const int direction);

#if 0
/* For future auto-solver */
extern void SolveSomeTiles(HexagonsWidget w);
#endif
extern void DrawAllTiles(HexagonsWidget w);
extern Boolean CheckSolved(HexagonsWidget w);
extern void InitMoves(void);
extern void PutMove(int direction);
extern void GetMove(int *direction);
extern int  MadeMoves(void);
extern void FlushMoves(HexagonsWidget w);
extern int  NumMoves(void);
extern void ScanMoves(FILE * fp, HexagonsWidget w, int moves);
extern void PrintMoves(FILE * fp);
extern void ScanStartPosition(FILE * fp, HexagonsWidget w);
extern void PrintStartPosition(FILE * fp, HexagonsWidget w);
extern void SetStartPosition(HexagonsWidget w);
extern int  Row(HexagonsWidget w, const int pos);
extern int  TrBl(HexagonsWidget w, const int pos, const int posRow);
extern int  TlBr(HexagonsWidget w, const int pos, const int posRow);

#endif /* _HexagonsP_h */
